######################################################################
###                                                                ###
### Linear Kalman FILTER                                           ###
###                                                                ###
###                                                                ###
######################################################################

# DESCRIPTION : Filtering function.
#----------------------------------
KF.VARG.with.Deltas <- function(Model,
                                A, B, D,
                                DATA,
                                res.4.cds.pricing=NaN,
                                mult.factor.delta) {
  
  
  # ===================================================================
  # ===================================================================
  # If res.4.cds.pricing==NaN then the EKF (CDS formula) is used
  # ===================================================================
  # ===================================================================
  
  
  # DIMENSIONS : 
  #=============
  # 1. parameters
  # Nu.tilde              : n * 1
  # Mu.tilde              : n * 1
  # Alpha.tilde           : n * 1
  # Beta.tilde            : n * n
  # 2. measurement param
  # observable      : m * Time
  # A               : m * 1
  # B               : n * m
  # D               : m * m
  # 4. initialization
  # W0              : n * 1
  # P0              : n * n
  
  #   save(n, W0, V0,
  #        Mu.tilde,
  #        Alpha.tilde,
  #        Nu.tilde,
  #        Beta.tilde,
  #        A, B, D,
  #        observable,
  #        res.4.cds.pricing,file="results/tempo.sav")
  

  n.e <- Model$n.e
  n.y <- Model$n.y
  
  observable <- rbind(t(DATA$CDS),
                      t(DATA$RF.yds),
                      matrix(0,1,dim(DATA$CDS)[1]),
                      mult.factor.delta * t(DATA$obs.delta))
  
  m <- nrow(observable)
  n <- Model$n.y + Model$n.e
  time <- ncol(observable)
  
  W.t.t <- matrix(0,n,time)
  W.t.t1 <- matrix(0,n,time)
  Y.t.t1 <- matrix(0,m,time)
  P.t.t <- array(0,c(n,n,time))
  P.t.t1 <- array(0,c(n,n,time))
  M.t.t1 <- array(0,c(m,m,time))
  loglik <- rep(0,time)
  
  # Measurement equation parameters
  #--------------------------------
  A <- matrix(A, nrow = m, ncol = time)
  B <- matrix(B, nrow = n, ncol = m)
  D <- matrix(D, nrow = m)
  
  moments <- Model$moments
  W0 <- moments$unc.mean
  W0[(n.y+1):(n.y+n.e)] <- 0 # All delta's are initially at zero
  V0 <- moments$unc.var
  if(is.na(V0[1])){
    V0 <- 10 * diag(c(W0)) #+ diag(n.y+n.e)
  }
  
  # Filter initialization
  W <- matrix(W0, ncol = 1)
  P <- matrix(V0, nrow = n, ncol = n)
  
  # Transition Equation parameters
  #-------------------------------
  
  # building the pre-multiplicative matrix:
  AR.mat <- moments$M1
  
  # Initialization of conditinal varianceof W:
  conVar <- matrix(0,n.y+n.e,n.y+n.e) # This is the average conditional covariance matrix
  
  for (t in 1:time){
    
    # mean prediction
    #-----------------
    W.pred <- moments$M0 + moments$M1 %*% W[1:n]
    
    # variance prediction
    #--------------------
    conVar[1:n.y,1:n.y] <- diag(c(moments$M2 + moments$M3 %*% W))
    conVar[(n.y+1):(n.y+n.e),(n.y+1):(n.y+n.e)] <- 
      moments$M4 %*% diag(c(moments$M2 + moments$M3 %*% W)) %*% t(moments$M4) +
      diag(c(moments$M5 + moments$M6 %*% W))
    conVar[(n.y+1):(n.y+n.e),1:n.y] <- 
      moments$M4 %*% diag(c(moments$M2 + moments$M3 %*% W))
    conVar[1:n.y,(n.y+1):(n.y+n.e)] <- 
      t(moments$M4 %*% diag(c(moments$M2 + moments$M3 %*% W)))
    
    P.pred <- AR.mat %*% P %*% t(AR.mat) + conVar
    
    # measurement equations
    #----------------------
    
    Y.pred <- A[,t] + t(B) %*% W.pred
    
    if(class(res.4.cds.pricing)=="list"){# In this case we use the CDS formula
      u <- -10000
      
      maturities.in.months <- 12 * DATA$maturities.in.years
      nb.mat <- length(maturities.in.months)
      
      res.pricing.CDS.single.date <- 
        pricing.CDS.single.date(res.4.cds.pricing$Model,u,res.4.cds.pricing$max.H,
                                matrix(W.pred,ncol=1),
                                res.4.cds.pricing$A.B.zeta.tilde,
                                res.4.cds.pricing$A.B.zeta.p.p.p,
                                res.4.cds.pricing$A.B.zeta.p.p,
                                res.4.cds.pricing$A.B.zeta.p,
                                fast.factor=fast.factor)
      CDS.based.on.W.pred <-
        c(t(120000*res.pricing.CDS.single.date$prices[2:n.e,maturities.in.months/fast.factor]))
      Y.pred[1:((n.e-1)*nb.mat)] <- CDS.based.on.W.pred
      
      grad.CDS.wrt.W <- 120000*res.pricing.CDS.single.date$grad.prices[2:n.e,,maturities.in.months/fast.factor]
      indices <- (1:(n.e-1))%x%rep(1,(n.y+n.e)*nb.mat) + rep(1,n.e-1) %x% ((n.e-1)*((1:((n.y+n.e)*nb.mat))-1))
      BB <- matrix(grad.CDS.wrt.W[indices],(n.y+n.e),(n.e-1)*nb.mat)
      
      B[,1:(nb.mat*(n.e-1))] <- BB
    }
    
    M.pred <- t(B) %*% P.pred %*% B + D %*% t(D)
    
    if(sum(is.na(M.pred))>0){
      return(list(
        loglik.vector = - 10^10
      ))
    }
    
    if ((det(M.pred))<1e-80){
      lik.val <- -1e8
      loglik[t] = lik.val
      break
    }
    
    residual  <- observable[,t] - Y.pred
    B.tilde <- B
    m.tilde <- m
    
    M.pred.reduced <- M.pred
    
    if(sum(is.na(observable[,t]))>0){
      indic.not.na <- which(!is.na(observable[,t]))
      m.tilde <- length(indic.not.na)
      #print(m.tilde)
      Y.pred.reduced <- matrix(Y.pred[indic.not.na], ncol = 1)
      M.pred.reduced <- M.pred[indic.not.na,]
      M.pred.reduced <- M.pred.reduced[,indic.not.na]
      residual <- observable[indic.not.na,t] - Y.pred.reduced
      B.tilde <- B[,indic.not.na]
    }
    
    AUX <- eigen(M.pred.reduced)
    if(max(abs(AUX$values))/min(abs(AUX$values))>10^50){
      inv.M.pred <- ginv(M.pred.reduced)
    }else{
      inv.M.pred <- solve(M.pred.reduced,tol=10^(-50))
    }

    Gain <- P.pred %*% B.tilde %*% inv.M.pred
    
    W.up <- W.pred + Gain %*% residual
    P.up <- (diag(1,n) - Gain %*% t(B.tilde)) %*% P.pred
    
    lik.val <- -.5*(m.tilde*log(2*pi) + log(det(M.pred.reduced)) + t(residual) %*% inv.M.pred %*% residual)
    
    W.up[W.up<0] <- 0
    
    
    W.t.t[,t] <- W.up
    W.t.t1[,t] <- W.pred
    Y.t.t1[,t] <- Y.pred
    P.t.t[,,t] <- P.up
    P.t.t1[,,t] <- P.pred
    M.t.t1[,,t] <- M.pred
    loglik[t] <- lik.val
    
    
    W <- W.up
    P <- P.up
    
  }
  
  # 4. Sending results back.
  #=========================
  
  result <- list(loglik.vector = loglik,
                 W.updated = W.t.t,
                 P.updated = P.t.t,
                 W.predicted = W.t.t1,
                 P.predicted = P.t.t1,
                 Y.predicted = Y.t.t1,
                 M.predicted = M.t.t1,
                 observable = observable
  )
  return(result)
}


Kalman_smoother <- function(Model,
                            A, B, D,
                            DATA,
                            res.4.cds.pricing=NaN,
                            mult.factor.delta){
  
  output_filter <- KF.VARG.with.Deltas(Model,
                                       A, B, D,
                                       DATA,
                                       res.4.cds.pricing,
                                       mult.factor.delta)
  rho_tt        <- t(output_filter$W.updated)
  Sigma_tt      <- output_filter$P.updated
  Sigma_tp1_t   <- output_filter$P.predicted
  rho_tp1_t     <- t(output_filter$W.predicted)
  Omega_tp1_t   <- output_filter$M.predicted
  
  H <- Model$moments$M1
  
  G <- t(B)
  
  # Number of observed variables: 
  ny = NROW(G)
  # Number of unobs. variables:
  nr = NCOL(G)
  # Number of time periods:
  TT = NROW(rho_tt)
  
  #Initilize output matrices:
  rho_tT       <- matrix(0,TT,nr)
  
  
  rho_tT[TT,]   <- rho_tt[TT,]
  Sigma_tT     <- array(0,c(nr,nr,TT))
  Sigma_tT[,,TT] <- Sigma_tt[,,TT]
  Var.model.implied.obs <- matrix(NaN,TT,ny) # this matrix will contain the variances of G*rho_t
  
  Var.model.implied.obs[TT,] <- diag(G %*% Sigma_tT[,,TT] %*% t(G)) + diag(D^2)
  
  for(t in seq(TT-1,1,by=-1)){
    
    AUX <- diag(Sigma_tt[,,t])
    AUX[AUX==0] <- .0000001
    diag(Sigma_tt[,,t]) <- AUX
    
    AUX <- diag(Sigma_tp1_t[,,t+1])
    AUX[AUX==0] <- .0000001
    diag(Sigma_tp1_t[,,t+1]) <- AUX
    
    F <- as.matrix(Sigma_tt[,,t] %*% t(H) %*% ginv(Sigma_tp1_t[,,t+1]))
    rho_tT[t,] <- rho_tt[t,] + t( F %*% (rho_tT[t+1,]-rho_tp1_t[t+1,]) )
    
    rho_tT[t,] = pmax(rho_tT[t,],0)
    
    Sigma_tT[,,t] <- c(Sigma_tt[,,t] + F %*%
                         (Sigma_tT[,,t+1]-Sigma_tp1_t[,,t+1]) %*% t(F))
    Var.model.implied.obs[t,] <- diag(
      G %*% Sigma_tT[,,t] %*% t(G)
    ) + diag(D^2)
  }
  #fitted.obs <- mu_t + rho_tT %*% t(G) # fitted observables
  output = list(W.smoothed=t(rho_tT),
                P.smoothed=Sigma_tT,
                Var.model.implied.obs=Var.model.implied.obs)
  return(output)
}


