

# ===============================================
# ======= ESTIMATION PROCEDURES =================
# ===============================================



loglik <- function(Theta,THETA,FILTER,DATA,kappa,weights.spds,
                   indic.cds=0,
                   indic.vec=0,
                   indic_max.avg.annual.SharpeRatio=NaN){
  # Computation of the log-likelihood function.
  # This function is the one used in the nuemrical optimization.
  
  theta <- THETA.2.theta(THETA)
  theta[FILTER==1] <- Theta
  new.THETA <- theta.2.THETA(theta)
  Model <- THETA.2.MODEL(new.THETA,kappa,weights.spds)
  
  res.filter <- run.Filter(Model,DATA,
                           indic.cds=indic.cds,
                           indic_max.avg.annual.SharpeRatio = indic_max.avg.annual.SharpeRatio)
  
  if(indic.vec==0){
    logl <- - sum(res.filter$loglik.vector)
    if(!is.na(indic_max.avg.annual.SharpeRatio)){
      # In this case, we add e penalty if (res.filter$max.Sharpe > indic_max.avg.annual.SharpeRatio)
      if(is.na(mean(res.filter$max.Sharpe.t))){
        penalty.term <- 100000000
      }else{
        penalty.term <- 10^5 * (mean(res.filter$max.Sharpe.t) > indic_max.avg.annual.SharpeRatio) * (mean(res.filter$max.Sharpe.t) - indic_max.avg.annual.SharpeRatio)
      }
      logl <- logl + c(penalty.term)
    }
  }else{
    logl <- res.filter$loglik.vector
  }
  return(logl)
}


Grad.loglik <- function(eps,Theta,THETA,FILTER,DATA,kappa,weights.spds,indic.cds=0){
  # This function computes the gradient of the log-likelihood, for each date.
  # That is, its output is a vector of dimension T x 1.
  
  if(sum(FILTER)!=length(Theta)){
    print("Inconsistency between FILTER and Theta's length.")
  }
  
  vec.grad.loglik <- NULL
  for(i in 1:length(Theta)){
    
    print(i)
    
    for(e.i in c(-eps[i],eps[i])){
      Theta.i <- Theta
      Theta.i[i] <- Theta.i[i] + e.i
      
      theta <- THETA.2.theta(THETA)
      theta[FILTER==1] <- Theta.i
      new.THETA <- theta.2.THETA(theta)
      Model <- THETA.2.MODEL(new.THETA,kappa,weights.spds)
      res.filter <- run.Filter(Model,DATA,
                               indic.cds=indic.cds)
      if(e.i==-eps[i]){
        logl.plus.eps.i <- res.filter$loglik.vector
      }else{
        logl.minus.eps.i <- res.filter$loglik.vector
      }
    }
    
    vec.grad.loglik <- cbind(vec.grad.loglik,(logl.plus.eps.i-logl.minus.eps.i)/(2*eps[i]))
  }
  
  return(vec.grad.loglik)
}



THETA.2.theta <- function(THETA){
  # This function converts the vector of parameters THETA into a vector theta of unconstrained parameters.
  # Typically, for those components of THETA that have to be non-negative, we have theta[i] = log(THETA[i]).
  
  if(THETA[6]==0){
    THETA[6] <- exp(-200)
  }
  theta <- log(THETA)
  theta[3] <- log(THETA[3]/.9999999/(1 - THETA[3]/.9999999))
  theta[7] <- log(THETA[7]/.9999999/(1 - THETA[7]/.9999999))
  theta[9] <- log(THETA[9]/.9999999/(1 - THETA[9]/.9999999))
  theta[17] <- log(THETA[17]/(1 - THETA[17]))
  theta[14] <- ifelse(THETA[14]==max.value.4.S,20,
                      log(THETA[14]/max.value.4.S/(1 - THETA[14]/max.value.4.S)))
  return(theta)
}

theta.2.THETA <- function(theta){
  # This function converts the vector of parameters theta into a vector THETA of constrained parameters.
  # Typically, for those components of THETA that have to be non-negative, we have theta[i] = log(THETA[i]).
  
  THETA <- exp(theta)
  THETA[3] <- .9999999 * exp(theta[3])/(1 + exp(theta[3])) # beta.y.y
  if(theta[6]< -100){
    THETA[6] <- 0
  }
  THETA[7]  <- .9999999 * exp(theta[7])/(1 + exp(theta[7])) # auto-regressive param of country-specific factors (same across entities)
  THETA[9]  <- .9999999 * exp(theta[9])/(1 + exp(theta[9])) # auto-regressive param of short-term rate
  THETA[17] <- exp(theta[17])/(1 + exp(theta[17]))
  THETA[14] <- max.value.4.S * exp(theta[14])/(1 + exp(theta[14])) # S
  return(THETA)
}


THETA.2.MODEL <- function(THETA,kappa,weights.spds){
  # This function converts the vector THETA of parameters into a Model.
  # A Model is a list object whose components are the scalara, vectors and matrices specifying the model.
  
  EFFECTIVE.KAPPA  <- kappa^THETA[17]
  EFFECTIVE.KAPPA <- EFFECTIVE.KAPPA / sum(EFFECTIVE.KAPPA)
  
  n.e <- length(kappa) + 1 # + 1 because of risk-free entity
  n.y <- 1 + (n.e-1) + 1 # common factor / country specific factors /short-term factor
  
  # y_t is organized as follows:
  # (common) y_t, x_t and r_t.
  
  # Specification of default intensities:
  beta.lambda <- c(0,THETA[(15+2*n.e+1):(15+2*n.e+(n.e-1))]*weights.spds)
  C <- c(0,THETA[(18+1):(18+(n.e-1))]*weights.spds)

  # Specification of common factor:
  beta.y.y <- THETA[3] # auto-regressive parameter of common factor
  nu_y <- THETA[4] # nu parameter for common factor
  I <- THETA[5] # indirect contagion parameter
  
  # Specification of country-specific factors:
  beta.x.y <- THETA[6] # influence of common factor on country-specific factors.
  beta.x.x <- THETA[7] # auto-correlation of country-specific factors (same across entities)
  
  nu_x <- THETA[18]
  
  # Specification of the risk-free short-term rate:
  mu.r <- THETA[8]
  rho.r <- THETA[9]
  beta.r <- rho.r/mu.r
  alpha.r <- THETA[10]/12 * (1 - beta.r * mu.r)/mu.r
  
  # SDF specification:
  theta.r <- THETA[11]
  theta.y <- THETA[12]
  theta.x <- THETA[13]
  S <- THETA[14] # credit-event-pricing parameter
  
  CDS.std.err.measur <- THETA[15]
  RF.std.err.measur <- THETA[16]
  
  # Convert what precedes into VARG specification

  alpha.y <- c(0,rep(0,n.e-1),alpha.r)
  beta.yy <- diag(c(beta.y.y,rep(beta.x.x,n.e-1),beta.r))
  beta.yy[2:n.e,1] <- beta.x.y
  
  beta.ye <- matrix(0,n.y,n.e) # this is beta_{e}^{(y)}, effect of delta on y
  # Same parameterisation of "contagiousness" as for matrix C:
  beta.ye[1,2:n.e] <- I * c(1,THETA[(18+(n.e-1)+1):(18+(n.e-1)+(n.e-1-1))])
  
  mu.y <- c(rep(1,n.y-1),mu.r)
  nu.y <- c(nu_y,rep(nu_x,n.y-2),0)
  
  # Specifcation of delta_t:
  alpha.e <- rep(0,n.e)
  beta.ey <- matrix(0,n.e,n.y) # this is beta_{y}^{(e)}, effect of y on delta
  diag(beta.ey) <- beta.lambda
  
  AUX <- c(0,1,THETA[(18+(n.e-1)+1):(18+(n.e-1)+(n.e-1-1))])
  AUX <- AUX/sum(AUX)
  beta.ee <- matrix(C,ncol=1) %*% matrix(AUX,nrow=1)

  # short-term risk-free rate as a function of w_t:
  xi.0 <- 0
  xi.1 <- rep(0,n.y+n.e)
  xi.1[n.y] <- 1
  
  # Recovery rate specification:
  a.0 <- rep(0,n.e)
  a.1 <- rbind(
    matrix(0,n.y,n.e),
    diag(n.e)
  )
  
  theta.priceofrisk.y <- c(theta.y,theta.x*EFFECTIVE.KAPPA,theta.r)
  theta.priceofrisk.delta <- S * c(0,EFFECTIVE.KAPPA)
  
  # Build a list that contains the parameterization of the physical dynamics:
  Model <- list(
    n.e = n.e,
    n.y = n.y,
    xi.0=xi.0,xi.1=xi.1,
    a.0=a.0, a.1=a.1,
    alpha.y=alpha.y,alpha.e=alpha.e,
    mu.y=mu.y,mu.e=mu.e,
    nu.y=nu.y,
    beta.yy=beta.yy,beta.ey=beta.ey,beta.ee=beta.ee,beta.ye=beta.ye,
    n.e = n.e,
    n.y = n.y,
    theta.priceofrisk.y = theta.priceofrisk.y,
    theta.priceofrisk.delta = theta.priceofrisk.delta,
    CDS.std.err.measur = CDS.std.err.measur,
    RF.std.err.measur = RF.std.err.measur
  )
  
  return(Model)
}


MODEL.2.THETA <- function(Model,kappa,weights.spds){
  # This function converts a Model into the vector THETA of parameters.
  # A Model is a list object whose components are the scalara, vectors and matrices specifying the model.
  
  n.e <- length(kappa) + 1
  n.y <- 1 + (n.e-1) + 1
  
  THETA <- matrix(0,15+2*n.e+(n.e-1))
  
  # y_t is organized as follows:
  # (common) y_t, x_t and r_t.
  
  # Specification of default intensities:
  beta.lambda <- diag(Model$beta.ey)
  THETA[(15+2*n.e+1):(15+2*n.e+(n.e-1))] <- beta.lambda[2:n.e]/weights.spds

  # Specification of the common factor:
  beta.y.y <- Model$beta.yy[1,1] # auto-regressive parameter of common factor
  THETA[3] <- beta.y.y
  THETA[4] <- Model$nu.y[1] # nu parameter for common factor
  
  aux.vec <- Model$theta.priceofrisk.delta[2:n.e]
  THETA[17] <- log(aux.vec[1]/aux.vec[2])/log(kappa[1]/kappa[2])
  EFFECTIVE.KAPPA <- kappa^THETA[17]
  EFFECTIVE.KAPPA <- EFFECTIVE.KAPPA / sum(EFFECTIVE.KAPPA)

  I <- Model$beta.ye[1,2]
  THETA[5] <- I # indirect contagion parameter
  
  # effect of country specific factors on associated lambda
  C <- apply(Model$beta.ee,1,function(x){sum(x,na.rm=TRUE)})
  C <- C[2:n.e]/weights.spds
  THETA[19:(18+(n.e-1))] <- C
  
  AUX <- Model$beta.ee[2,2:n.e]/(C[2]*weights.spds[2])
  AUX <- AUX/AUX[1]
  THETA[(18+(n.e-1)+1):(18+(n.e-1)+(n.e-1-1))] <- AUX[2:(n.e-1)]

  # Specification of country-specific factors:
  beta.x.y <- Model$beta.yy[2,1]
  THETA[6] <- beta.x.y # influence of common factor on country-specific factors.
  beta.x.x <- Model$beta.yy[2,2]
  THETA[7] <- beta.x.x # auto-regressuence of country-specific factors (same across entities)
  THETA[18] <- Model$nu.y[2] # nu_x

  # Specification of the risk-free short-term rate:
  
  mu.r <- Model$mu.y[n.y]
  THETA[8] <- mu.r
  beta.r <- Model$beta.yy[n.y,n.y]
  THETA[9] <-  beta.r * mu.r
  alpha.r <- Model$alpha.y[n.y]
  THETA[10] <- 12 * alpha.r * mu.r / (1 - beta.r * mu.r)
  
  # SDF specification:
  
  theta.y <- Model$theta.priceofrisk.y[1]
  theta.x <- max(Model$theta.priceofrisk.y[2:n.e]/EFFECTIVE.KAPPA,na.rm=TRUE)
  theta.r <- Model$theta.priceofrisk.y[n.y]
  
  THETA[11] <- theta.r
  THETA[12] <- theta.y
  THETA[13] <- theta.x
  
  S <- max(Model$theta.priceofrisk.delta[2:n.e]/EFFECTIVE.KAPPA,na.rm=TRUE)
  THETA[14] <- S # credit-event-pricing parameter
  
  THETA[15] <- Model$CDS.std.err.measur
  THETA[16] <- Model$RF.std.err.measur
  
  THETA[THETA==0] <- exp(-100)

  return(THETA)
}


g.complete.new <- function(THETA,kappa,weights.spds,ctry.names){
  # Function that converts estimated param into param of interest
  # (that will appear in the parameter table)
  
  Model <- THETA.2.MODEL(THETA,kappa,weights.spds)
  
  n.e <- Model$n.e
  n.y <- Model$n.y
  
  ell <- THETA[17]
  
  beta.lambda <- THETA[(15+2*n.e+1):(15+2*n.e+(n.e-1))]*weights.spds
  C <- THETA[(18+1):(18+(n.e-1))]*weights.spds

  AUX <- c(1,THETA[(18+(n.e-1)+1):(18+(n.e-1)+(n.e-1-1))])
  AUX.4.beta.ee <- AUX/sum(AUX)

  latex.names.beta.lambda.x <- NULL
  latex.names.C <- NULL
  latex.names.kappa.C <- NULL
  for(ctry in ctry.names){
    eval(parse(text = gsub(" ","",paste("latex.names.beta.lambda.x <- c(latex.names.beta.lambda.x,\"$\\\\beta^{(x)}_{\\\\lambda,",ctry,"}$\")",
                                        sep=""))))
    eval(parse(text = gsub(" ","",paste("latex.names.C <- c(latex.names.C,\"$c_{",ctry,"}$\")",
                                        sep=""))))
    eval(parse(text = gsub(" ","",paste("latex.names.kappa.C <- c(latex.names.kappa.C,\"$\\\\boldsymbol\\\\kappa_{c,",ctry,"}$\")",
                                        sep=""))))
  }
  
  # Prices of risk:
  # SDF specification:
  theta.y <- THETA[12]
  theta.x <- THETA[13]
  theta.r <- THETA[11]
  S <- THETA[14] # credit-event-pricing parameter
  
  vec.param <- c(
    Model$alpha.y[n.y], # [1]
    Model$mu.y[n.y], # [2]
    Model$beta.yy[n.y,n.y], # [3]
    1-Model$beta.yy[1,1], # [4]
    1-Model$beta.yy[2,2], # [5]
    Model$beta.yy[2,1], # [6]
    Model$nu.y[1], # [7] nu_z
    Model$nu.y[2], # [8] nu_x
    beta.lambda, # [9:(7+n.e)]
    C, # [(8+n.e):(6+2*n.e)]
    AUX.4.beta.ee, # [(7+2*n.e):(5+3*n.e)]
    ell, # [6+3*n.e]
    theta.y, # [7+3*n.e]
    theta.x, # [8+3*n.e]
    theta.r, # [9+3*n.e]
    S, # [10+3*n.e]
    Model$RF.std.err.measur, # [11+3*n.e]
    Model$CDS.std.err.measur # [12+3*n.e]
  )
  
  names.param <- c(
    "$\\alpha_r$",
    "$\\mu_r$",
    "$\\beta_r$",
    "$1-\\beta_z^{(z)}$",
    "$1-\\beta_x^{(x)}$",
    "$\\beta_x^{(z)}$",
    "$\\nu_z$",
    "$\\nu_x$",
    latex.names.beta.lambda.x,
    latex.names.C,
    latex.names.kappa.C,
    "$\\ell$",
    "$\\theta_z$",
    "$\\theta_x$",
    "$\\theta_r$",
    "$\\mathbf{S}$",
    "$\\sigma_{RF}$",
    "$\\eta_{CDS}$"
  )
  return(list(vec.param=vec.param,names.param=names.param))
}




run.Filter <- function(Model,DATA,
                       indic.cds=0, # 1 to use CDS formula (RMV otherwise)
                       indic.smoother = 0,
                       mult.factor.delta = 100,
                       indic_max.avg.annual.SharpeRatio = NaN # If = NaN, nothing; if = to a scalar, this scalar is the max of the max Sharpe ratios
){
  # The aim of this function is to call the Extended Kalman Filter on given (a) Model and (b) dataset.
  
  n.e <- Model$n.e
  n.y <- Model$n.y
  
  Model.with.moments <- Model
  Model.with.moments$moments <- compute.moments(Model)
  W0 <- Model.with.moments$moments$unc.mean
  
  Mu <- c(Model$mu.y,
          Model$mu.e)
  Beta <- t(rbind(
    cbind(Model$beta.yy,Model$beta.ye),
    cbind(Model$beta.ey,Model$beta.ee)
  )) # The columns give the effect of w on (y,e)
  Alpha <- c(Model$alpha.y,Model$alpha.e)
  Nu <- c(Model$nu.y,rep(0,n.e))
  
  PHI <- (Mu %*% matrix(1, nrow = 1, ncol = n.y+n.e)) * t(Beta)
  
  n.y <- Model$n.y
  n.e <- Model$n.e
  
  Model.Q <- make.Model.Q(Model)
  
  max.H <- max(12*DATA$maturities.in.years)
  nb.mat <- length(DATA$maturities.in.years)
  
  maturities.in.months <- 12*DATA$maturities.in.years
  
  rmv.pricing.Q <- pricing.RMV(Model.Q,maturities.in.months,W=NULL)
  
  # ===== the Constant in the price formula
  # For all countries except baseline:
  a.h <- rmv.pricing.Q$A.B$B[,2:n.e] - matrix(rmv.pricing.Q$A.B$B[,1],nb.mat,n.e-1)
  aux <- -120000/matrix(maturities.in.months,nb.mat,n.e-1)
  a.h <- c(a.h * aux)
  
  # ===== the Slopes in the price formula
  # For all countries except baseline:
  # Remark: in rmv.pricing.Q$A.B$A:
  #   - nb of rows = nb of pricing factors,
  #   - nb of columns = nb of entities (including the risk-free one),
  #   - nb of layers = nb of maturities.
  b.h <- rmv.pricing.Q$A.B$A[,2:n.e,] - array(rmv.pricing.Q$A.B$A[,1,] %x% matrix(1,1,n.e-1),c(n.y+n.e,n.e-1,nb.mat))
  aux <- -120000/array((maturities.in.months)%x%rep(1,(n.y+n.e)*(n.e-1)),c(n.y+n.e,n.e-1,nb.mat))
  b.h <- aux * b.h
  b.h <- matrix(b.h,n.y+n.e,(n.e-1)*nb.mat)
  incr <- (1:(n.e-1)) %x% rep(1,nb.mat)
  index.column <- rep(1,n.e-1) %x% seq(0,(n.e-1)*nb.mat-1,by=n.e-1) + incr
  b.h.delta <- b.h[(n.y+1):(n.y+n.e),index.column]
  b.h <- b.h[1:n.y,index.column]
  
  if(indic.cds==1){# In this case, we will use the RFV-CDS formula associated with the EKF.
    u <- -10000
    res.4.cds.pricing <- pricing.CDS(Model.Q,u,
                                     max(maturities.in.months),matrix(c(W0[1:n.y],rep(0,n.e)),ncol=1))
  }else{# if res.4.cds.pricing=NaN then the KF will not be "extended" 
    res.4.cds.pricing <- NaN
  }
  
  # Credit-Riskfree yields:
  a.RF.Q <- -1200/maturities.in.months*rmv.pricing.Q$A.B$B[,1]
  b.RF.Q <- -1200/t(matrix(maturities.in.months,nb.mat,n.y))*
    rmv.pricing.Q$A.B$A[1:n.y,1,]
  b.RF.Q.delta <- -1200/t(matrix(maturities.in.months,nb.mat,n.e))*
    rmv.pricing.Q$A.B$A[(n.y+1):(n.y+n.e),1,]
  
  a.h <- c(a.h,a.RF.Q)
  b.h <- cbind(b.h,b.RF.Q)
  b.h.delta <- cbind(b.h.delta,b.RF.Q.delta)
  b.h <- rbind(b.h,
               b.h.delta)
  
  observables <- rbind(t(DATA$CDS),
                       t(DATA$RF.yds))
  
  D.matrix <- diag(
    c(Model$CDS.std.err.measur * apply(DATA$CDS,2,function(x){sd(x,na.rm=TRUE)}),
      rep(Model$RF.std.err.measur,nb.mat)))
  
  # Convert the vector of constant in the measurement equations into a matrix:
  a.h <- matrix(a.h,length(a.h),dim(observables)[2])

  # Add parameters for measurement equations associated with delta's:
  a.h <- rbind(a.h,matrix(0,n.e,dim(observables)[2]))
  b.h <- cbind(b.h,
               rbind(matrix(0,n.y,n.e),
                     mult.factor.delta*diag(n.e)))
  D.matrix.extended <- mult.factor.delta * .01 * diag(dim(D.matrix)[1]+n.e)
  D.matrix.extended[1:dim(D.matrix)[1],1:dim(D.matrix)[1]] <- D.matrix
  
  # Launching the Kalman Filter
  filtered.values <- KF.VARG.with.Deltas(Model.with.moments,
                                         a.h, b.h, D.matrix.extended,
                                         DATA,
                                         res.4.cds.pricing=res.4.cds.pricing,
                                         mult.factor.delta)
  
  if(indic.smoother==1){
    res.smoother <- Kalman_smoother(Model.with.moments,
                                    a.h, b.h, D.matrix.extended,
                                    DATA,
                                    res.4.cds.pricing=res.4.cds.pricing,
                                    mult.factor.delta)
    filtered.values$W.smoothed <- res.smoother$W.smoothed
    filtered.values$P.smoothed <- res.smoother$P.smoothed
  }
  
  
  # Sharpe Ratios:
  
  if(!is.na(indic_max.avg.annual.SharpeRatio)){
    price.of.risk <- c(Model$theta.priceofrisk.y,
                       Model$theta.priceofrisk.delta)
    
    # First step, compute mu.1.m and mu.2.m (see Online Appendix for notations)
    # SDF: M_{t,t+1} = exp(mu.0.m + mu.1.m'w_{t+1} + mu.2.m'w_{t})
    
    U.0 <- array(c(price.of.risk),
                 c(n.y+n.e,1,1))
    
    res <- compute.AB(Model$beta.yy,Model$beta.ye,Model$beta.ey,Model$beta.ee,
                      Model$mu.y,Model$mu.e,
                      Model$alpha.y,Model$alpha.e,
                      Model$nu.y,
                      U.0,1)
    
    mu.1.m <- price.of.risk
    mu.2.m <- - Model$xi.1 - c(res$A)
    
    # Second step: compute E(M_{t,t+h}^2) and E(M_{t,t+h})
    U <- array(NaN,c(n.y+n.e,2,2))
    # for M^2:
    U[,1,1] <- 2 * mu.1.m
    U[,1,2] <- 2 * (mu.1.m + mu.2.m)
    # for M:
    U[,2,1] <- 1 * mu.1.m
    U[,2,2] <- 1 * (mu.1.m + mu.2.m)
    
    res <- compute.AB(Model$beta.yy,Model$beta.ye,Model$beta.ey,Model$beta.ee,
                      Model$mu.y,Model$mu.e,
                      Model$alpha.y,Model$alpha.e,
                      Model$nu.y,
                      U,c(1,12))

    # Evaluate maximum Sharpe ratio at unconditional mean:
    
    Z.0 <- Model.with.moments$moments$unc.mean
    Z.t <- filtered.values$W.updated
    
    matur <- 2 # select the 12-month maturity (number of entry of maturity given in compute.AB)
    
    if(is.null(Z.0)|is.null(Z.t)){
      filtered.values$max.Sharpe <- 1000
    }else{
      # Evaluated at unconditional mean:
      Psi.2.theta <- res$B[matur,1] + t(Z.0) %*% matrix(res$A[,1,matur],ncol=1)
      Psi.theta   <- res$B[matur,2] + t(Z.0) %*% matrix(res$A[,2,matur],ncol=1)
      filtered.values$max.Sharpe.0 <- sqrt(exp(Psi.2.theta - 2*Psi.theta)-1)
      
      # Time series of max conditional Sharpe ratios:
      Psi.2.theta <- res$B[matur,1] + t(Z.t) %*% matrix(res$A[,1,matur],ncol=1)
      Psi.theta   <- res$B[matur,2] + t(Z.t) %*% matrix(res$A[,2,matur],ncol=1)
      filtered.values$max.Sharpe.t <- sqrt(exp(Psi.2.theta - 2*Psi.theta)-1)
    }
    
  }else{
    # we do not consider maximum Sharpe ratio in this case
    filtered.values$max.Sharpe.0 <- NaN
    filtered.values$max.Sharpe.t <- NaN
  }

  return(filtered.values)
}



make.quanto.Model <- function(PARAM,Model.with.FX,W,delta.s,DATA){
  # Note: W is not of the right dimension (y_{s,t} not there)
  
  Model.with.FX.new <- Model.with.FX
  
  n.y <- Model.with.FX$n.y
  old.n.y <- n.y - 1 # for the model without FX
  n.e <- Model.with.FX$n.e
  
  param <- exp(PARAM)
  
  if(length(param)==n.e-1){
    # In this case, we estimate depreciations at default as well as theta.s
    a.w.s <- c(rep(0,old.n.y),
               1, # This is for y_{s,t}
               0, # delta risk-free component
               param[1:(n.e-1)] # depreciations at default
    )
    theta.s <- 0
  }
  if(length(param)==n.e){
    # In this case, we estimate depreciations at default as well as theta.s
    a.w.s <- c(rep(0,old.n.y),
               1, # This is for y_{s,t}
               0, # delta risk-free component
               param[1:(n.e-1)] # depreciations at default
    )
    theta.s <- param[n.e]
  }
  if(length(param)==2*n.e-1){
    # Estimate complete vector a.w.s (y,x,delta)
    
    # remove potential impact of deltas': 
    a.w.s <- c(param[1:n.e],# common factor + country-specific factors
               0,# short-term rate
               1,# y_{s,t}
               0,# delta risk-free entity
               indic.allow.4.u.delta*param[(n.e+1):(2*n.e-1)])
    theta.s <- 0
  }
  if(length(param)==2*n.e){
    # Estimate complete vector a.w.s (y,x,delta)
    
    # remove potential impact of deltas': 
    a.w.s <- c(param[1:n.e],# common factor + country-specific factors
               0,# short-term rate
               1,# y_{s,t}
               0,# delta risk-free entity
               indic.allow.4.u.delta*param[(n.e+1):(2*n.e-1)])
    theta.s <- param[2*n.e]
  }
  
  Model.with.FX.new$a.w.s <- a.w.s 
  
  Model.with.FX.new$theta.priceofrisk.y[n.y] <- theta.s
  
  # Estimation of y_{s,t} dynamics
  s.part.y <- t(W[1:(n.y-1),2:dim(W)[2]]) %*% matrix(a.w.s[1:(n.y-1)],ncol=1)
  y.s <- delta.s - a.0.s - s.part.y
  W.FX <- rbind(W[1:old.n.y,],c(E,y.s),
                W[(old.n.y+1):(old.n.y+n.e),])
  W.FX[n.y,] <- c(y.s[1],y.s)
  s.0 <- y.s[2:length(y.s)]
  s.1 <- y.s[1:(length(y.s)-1)]
  C <- var(cbind(s.0,s.1))
  correl <- max(C[1,2]/C[1,1],0)
  E <- - a.0.s
  V <- var(y.s)
  
  mu.s <- V/E
  nu.s <- E/mu.s
  beta.s <- 0
  
  # Update model accordingly:
  Model.with.FX.new$mu.y[n.y] <- mu.s
  Model.with.FX.new$nu.y[n.y] <- nu.s
  Model.with.FX.new$beta.yy[n.y,n.y] <- beta.s
  
  # Pricing of foreign-currency denotminated CDS:
  cds.pricing.Q.FX <- pricing.CDS(make.Model.Q(Model.with.FX.new),u,
                                  max(12*DATA$maturities.in.years),W.FX)
  
  return(list(
    cds.pricing.Q.FX = cds.pricing.Q.FX,
    Model.with.FX = Model.with.FX.new,
    W.FX = W.FX
  ))
}


dev.quanto.4.optim <- function(PARAM,
                               Model.with.FX,W,delta.s,DATA,
                               cds.pricing.Q,quanto.CDS.obs){
  
  res.AUX <- make.quanto.Model(PARAM,Model.with.FX,W,delta.s,DATA)
  
  quanto.CDS.Q <- 120000*(res.AUX$cds.pricing.Q.FX$prices-cds.pricing.Q$prices)
  quanto.CDS.Q.model <- quanto.CDS.Q[2:n.e,,12*DATA$maturities.in.years]
  
  pricing.errors <- quanto.CDS.Q.model - quanto.CDS.obs
  
  sum.squares.obs <- apply(quanto.CDS.obs,c(1,3),function(x){sum(x^2,na.rm=TRUE)})
  sum.squares.pricing.errors <- apply(pricing.errors,c(1,3),function(x){sum(x^2,na.rm=TRUE)})
  
  res <- 1000*sum(sum.squares.pricing.errors/sum.squares.obs) +
    10^10*sum(is.na(quanto.CDS.Q.model))
  
  return(res)
}




# ===============================================
# ======= PRICING PROCEDURES ====================
# ===============================================




compute.B <- function(Model,u){
  # This function contributes to the computation of the Laplace transform of
  # the state vector w_t.
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  n.e <- dim(beta.ey)[1]
  n.y <- dim(beta.ey)[2]
  
  u.y <- u[1:n.y]
  u.e <- c(0,u[n.y + (1:(n.e-1))]) # reincorporate the element corresponding to the credit-risk-free entity
  
  z.e <- u.e * mu.e / (1 - u.e * mu.e)
  z.y <- (u.y + t(beta.ey) %*% z.e) * mu.y / (1 - (u.y + t(beta.ey) %*% z.e) * mu.y)
  
  B.u <- matrix(alpha.y,nrow=1) %*% z.y + matrix(alpha.e,nrow=1) %*% z.e - 
    matrix(nu.y,nrow=1) %*% log(1 - (u.y + t(beta.ey) %*% z.e) * mu.y)
  return(B.u)
}



compute.A <- function(Model,u){
  # This function contributes to the computation of the Laplace transform of
  # the state vector w_t.
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  n.e <- dim(beta.ey)[1]
  n.y <- dim(beta.ey)[2]
  
  M <- matrix(0,n.e+n.y,n.e+n.y)
  M[1:n.y,1:n.y] <- t(beta.yy)
  M[n.y+(1:n.e),1:n.y] <- t(beta.ye)
  M[n.y+(1:n.e),n.y+(1:n.e)] <- t(beta.ee)
  
  u.y <- u[1:n.y]
  u.e <- c(0,u[n.y + (1:(n.e-1))]) # reincorporate the element corresponding to the credit-risk-free entity
  
  z.e <- u.e * mu.e / (1 - u.e * mu.e)
  z.y <- (u.y + t(beta.ey) %*% z.e) * mu.y / (1 - (u.y + t(beta.ey) %*% z.e) * mu.y)
  
  z.w <- c(z.y,z.e)
  A.u <- M %*% z.w
  
  A.u <- A.u[c(
    1:n.y,
    (n.y+2):(n.y+n.e)
  )]
  return(A.u)
}

compute.grad.A <- function(Model,u){
  # This function contributes to the computation of the gradient of the Laplace transform of
  # the state vector w_t.

  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  n.e <- dim(beta.ey)[1]
  n.y <- dim(beta.ey)[2]
  
  M <- matrix(0,n.e+n.y,n.e+n.y)
  M[1:n.y,1:n.y] <- t(beta.yy)
  M[n.y+(1:n.e),1:n.y] <- t(beta.ye)
  M[n.y+(1:n.e),n.y+(1:n.e)] <- t(beta.ee)
  
  u.y <- u[1:n.y]
  u.e <- c(0,u[n.y + (1:(n.e-1))]) # reincorporate the element corresponding to the credit-risk-free entity
  
  grad.A.u <- matrix(0,n.e+n.y,n.e+n.y)
  grad.A.u[n.y+(1:n.e),n.y+(1:n.e)] <- diag(c(mu.e * (1/(1 - u.e * mu.e)^2)))
  z.e <- u.e * mu.e / (1 - u.e * mu.e)
  z.y <- u.y + t(beta.ey) %*% z.e 
  grad.A.u[1:n.y,1:n.y] <- diag(c(mu.y * (1/(1 - z.y)^2)))
  grad.A.u[1:n.y,n.y+(1:n.e)] <-
    diag(c(mu.y * (1/(1 - z.y)^2))) %*% t(beta.ey) %*% diag(c(mu.e * (1/(1 - u.e * mu.e)^2)))
  
  grad.A.u <- M %*% grad.A.u
  
  # remove delta.0 (credit risk free):
  grad.A.u <- grad.A.u[,c(1:n.y,(n.y+2):(n.e+n.y))]
  grad.A.u <- grad.A.u[c(1:n.y,(n.y+2):(n.e+n.y)),]
  
  return(grad.A.u)
}



simul.model <- function(Alpha,
                        Beta,
                        Mu,
                        Nu,
                        nb.sim,
                        X0) {
  # This function allows to simulate the model, wirtten in a Vectorial AutoRegressive Gamma format.
  # X0 is the initial condition of the state vector.
  
  # Taking sizes
  n <- length(Alpha)
  Alpha <- matrix(Alpha, ncol = 1)
  Beta <- matrix(Beta, nrow = n, ncol = n)
  Mu <- matrix(Mu, ncol = 1)
  Nu <- matrix(Nu, ncol = 1)
  
  # Initialization
  X <- matrix(0, nrow = n, ncol = nb.sim)
  
  # Simulations of factors
  X[,1] <- c(X0)
  for (i in 2:nb.sim){
    param.pois <- rpois(n, Alpha + t(Beta) %*% X[,(i-1)])
    X[,i] <- rgamma(n,shape = param.pois + Nu, scale = Mu)
    
  }
  
  return(X)
}


compute.moments <- function(Model,indic.compute.uncvar=0){
  # This function computes M0 and M1 that are such that:
  #  E([y_t',delta_t']|F_{t-1}) = M0 + M1 [y_{t-1}',delta_{t-1}']
  # (see Online Appendix A.2 for notations)

  n.y <- Model$n.y
  n.e <- Model$n.e
  
  M0 <- matrix(c(Model$mu.y,Model$mu.e),ncol=1) * rbind(
    matrix(Model$nu.y + Model$alpha.y,ncol=1),
    matrix(Model$alpha.e + Model$beta.ey %*% (Model$mu.y * (Model$nu.y + Model$alpha.y)))
  )
  
  M1 <- diag(c(Model$mu.y,Model$mu.e)) %*% rbind(
    cbind(Model$beta.yy,Model$beta.ye),
    cbind(
      Model$beta.ey %*% (matrix(Model$mu.y,n.y,n.y) * Model$beta.yy),
      Model$beta.ey %*% (matrix(Model$mu.y,n.y,n.e) * Model$beta.ye) + Model$beta.ee
    )
  )
  
  unc.mean <- solve(diag(n.e+n.y) - M1) %*% M0
  
  M2 <- Model$mu.y * Model$mu.y * (Model$nu.y + 2*Model$alpha.y)
  
  M3 <- 2 * diag(c(Model$mu.y^2)) %*% cbind(Model$beta.yy,Model$beta.ye)
  
  M4 <- diag(c(Model$mu.e)) %*% Model$beta.ey
  
  M5 <- 2 * Model$mu.e^2 * Model$alpha.e + 2 * diag(c(Model$mu.e^2)) %*% Model$beta.ey %*%
    (Model$mu.y * (Model$nu.y + Model$alpha.y))
  
  M6 <- 2 * diag(c(Model$mu.e^2)) %*% cbind(
    Model$beta.ey %*% diag(c(Model$mu.y)) %*% Model$beta.yy,
    Model$beta.ey %*% diag(c(Model$mu.y)) %*% Model$beta.ye + Model$beta.ee
  )
  
  Sigma.E.y.delta <- matrix(0,n.y+n.e,n.y+n.e) # This is the average conditional covariance matrix
  Sigma.E.y.delta[1:n.y,1:n.y] <- diag(c(M2 + M3 %*% unc.mean))
  Sigma.E.y.delta[(n.y+1):(n.y+n.e),(n.y+1):(n.y+n.e)] <- M4 %*% diag(c(M2 + M3 %*% unc.mean)) %*% t(M4) +
    diag(c(M5 + M6 %*% unc.mean))
  Sigma.E.y.delta[(n.y+1):(n.y+n.e),1:n.y] <- M4 %*% diag(c(M2 + M3 %*% unc.mean))
  Sigma.E.y.delta[1:n.y,(n.y+1):(n.y+n.e)] <- t(M4 %*% diag(c(M2 + M3 %*% unc.mean)))
  
  if(indic.compute.uncvar==1){
    unc.var <- solve(diag((n.y + n.e)^2) - M1 %x% M1) %*% matrix(Sigma.E.y.delta,ncol=1)
    unc.var <- matrix(unc.var,n.y + n.e,n.y + n.e)
  }else{
    unc.var <- NaN
  }
  
  return(list(
    M0 = M0,
    M1 = M1,
    M2 = M2,
    M3 = M3,
    M4 = M4,
    M5 = M5,
    M6 = M6,
    unc.mean = unc.mean,
    unc.var = unc.var
  ))
}


a.function <- function(beta,mu,u){
  # Computation of the Laplace transform.
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # u is od dimension n.l x q
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  q <- dim(u)[2] # number of vectors u we simultaneously consider
  mu.q <- matrix(mu,ncol=1) %*% rep(1,q)
  res <- t(beta) %*% ( (u * mu.q)/(1 - u * mu.q) )
  return(res)
}

b.function <- function(beta,mu,alpha,nu,u){
  # Computation of the Laplace transform.
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is od dimension n.l x q
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  q <- dim(u)[2] # number of vectors u we simultaneously consider
  nu.l <- dim(beta)[1]
  nu.m <- dim(beta)[2]
  mu.q <- matrix(mu,ncol=1) %*% rep(1,q)
  aux <- 1 - u * mu.q
  aux[aux<0] <- NaN
  aux[which(nu!=0),] <- log(aux[which(nu!=0),])
  res <- matrix(alpha,nrow=1) %*% ( (u * mu.q)/(1 - u * mu.q) ) - matrix(nu,nrow=1) %*% aux
  return(res)
}

phi <- function(beta.yy,beta.ye,beta.ey,beta.ee,
                mu.y,mu.e,
                alpha.y,alpha.e,
                nu.y,
                u){
  # Computation of the Laplace transform.
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is of dimension n.l x q x 2
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  # The first "layer" of u corresponds to u_1, the second to u_2.
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.y <- matrix(u[1:n.y,],nrow=n.y)
  u.e <- matrix(u[(n.y+1):(n.e+n.y),],nrow=n.e)
  A <- rbind(
    a.function(beta=beta.yy,mu=mu.y,u = u.y + 
                 a.function(beta=beta.ey,mu=mu.e,u = u.e)
    ),
    a.function(beta=beta.ye,mu=mu.y,u = u.y + 
                 a.function(beta=beta.ey,mu=mu.e,u = u.e)) + a.function(beta=beta.ee,mu=mu.e,u = u.e)
  )
  B <- b.function(beta=beta.ye,mu=mu.y,alpha=alpha.y,nu=nu.y,u = u.y + 
                    a.function(beta=beta.ey,mu=mu.e,u = u.e)) +
    b.function(beta=beta.ee,mu=mu.e,alpha=alpha.e,nu=0*alpha.e,u = u.e)
  return(list(A=A,B=B))
}

compute.AB <- function(beta.yy,beta.ye,beta.ey,beta.ee,
                       mu.y,mu.e,
                       alpha.y,alpha.e,
                       nu.y,
                       u,HH.aux){
  # Computation of the Laplace transform.
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is of dimension n.l x q x 2
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  # The first "layer" of u corresponds to u_1, the second to u_2.
  # HH gives the considered horizons. It is either a vector or an integer that gives the maximum horizon considered
  
  if(length(HH.aux)==1){
    max.H <- HH.aux
    HH <- 1:HH.aux
  }else{
    max.H <- max(HH.aux)
    HH <- HH.aux
  }
  nb.mat <- length(HH)
  
  q <- dim(u)[2]
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  all.A <- array(NaN,c(n.y+n.e,q,nb.mat))
  all.B <- matrix(NaN,nb.mat,q)
  A <- matrix(0,n.y+n.e,q)
  B <- matrix(0,1,q)
  count <- 0
  for(i in 1:max.H){
    if(i==1){
      u.i <- u[,,1]
    }else{
      u.i <- u[,,2]
    }
    A.B <- phi(beta.yy,beta.ye,beta.ey,beta.ee,
               mu.y,mu.e,
               alpha.y,alpha.e,
               nu.y,
               u.i + A)
    A <- A.B$A
    B <- A.B$B + B
    if(sum(i==HH)==1){
      count <- count + 1
      all.A[,,count] <- A
      all.B[count,]  <- B
    }
  }
  return(list(A=all.A,B=all.B))
}


compute.proba.def <- function(Model,max.H,W){
  # This function computes the probabilities of default for the n.e entities and
  # all horizons in [1.max.H]
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- 0*Model$xi.0
  xi.1 <- 0*Model$xi.1
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.1 <- - rbind(
    matrix(0,n.y,n.e),
    100000*diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  u <- array(0,c(n.y+n.e,n.e,2))
  u[,,1] <- u.1
  u[,,2] <- u.2
  A.B <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                    mu.y,mu.e,
                    alpha.y,alpha.e,
                    nu.y,
                    u,max.H)
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  for(i in 1:max.H){
    for(e in 1:n.e){
      all.prices[e,,i] <- exp(matrix(A.B$A[,e,i] - xi.1,nrow=1) %*% W + A.B$B[i,e] - i*xi.0)
    }
  }
  
  return(1 - all.prices)
}



pricing.RMV <- function(Model,HH.aux,W){
  # This function computes prices and rates of bonds under the RMV assumptions
  
  if(length(HH.aux)==1){
    max.H <- HH.aux
    HH <- 1:HH.aux
  }else{
    max.H <- max(HH.aux)
    HH <- HH.aux
  }
  nb.mat <- length(HH)
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.1 <- - rbind(
    matrix(0,n.y,n.e),
    diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  u <- array(0,c(n.y+n.e,n.e,2))
  u[,,1] <- u.1
  u[,,2] <- u.2
  A.B <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                    mu.y,mu.e,
                    alpha.y,alpha.e,
                    nu.y,
                    u,HH.aux)
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],nb.mat))
  all.rates <- array(NaN,c(n.e,dim(W)[2],nb.mat))
  for(i in 1:nb.mat){
    for(e in 1:n.e){
      if(!is.null(W)){
        all.prices[e,,i] <- exp(matrix(A.B$A[,e,i] - xi.1,nrow=1) %*% W + A.B$B[i,e] - HH[i]*xi.0)
        all.rates[e,,i] <- -1/HH[i] * log(all.prices[e,,i])
      }
      A.B$A[,e,i] <- A.B$A[,e,i] - xi.1
      A.B$B[i,e]  <- A.B$B[i,e] - HH[i]*xi.0
    }
  }
  return(list(prices=all.prices,rates=all.rates,A.B=A.B))
}


pricing.RFV <- function(Model,u,max.H,W,indic.compute.deriv=0){
  # This function computes prices and rates of bonds under the RMV assumptions
  # u has to -> - infty
  # If indic.compute.deriv==1, the derivative of prices and rates will be computed
  #          w.r.t. the first column of W ONLY
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  n.w <- n.e + n.y
  
  # for zeta''':
  u.1 <- rbind(
    matrix(0,n.y,n.e),
    u * diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  U <- array(0,c(n.y+n.e,n.e,2))
  U[,,1] <- u.1
  U[,,2] <- u.2
  
  A.B.zeta.p.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  # for zeta'':
  u.1 <- u.1 - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                             mu.y,mu.e,
                             alpha.y,alpha.e,
                             nu.y,
                             U,max.H)
  
  # for zeta':
  u.1 <- - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                           mu.y,mu.e,
                           alpha.y,alpha.e,
                           nu.y,
                           U,max.H)
  
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  
  zeta.p.p.p <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p   <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p     <- array(NaN,c(n.e,dim(W)[2],max.H))
  aux <- 0
  for(i in 1:max.H){
    for(e in 1:n.e){
      zeta.p.p.p[e,,i] <- exp(          - i*xi.0 + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e,,i]   <- exp( - a.0[e] - i*xi.0 + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% W + A.B.zeta.p.p$B[i,e])
      zeta.p[e,,i]     <- exp( - a.0[e] - i*xi.0 + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% W + A.B.zeta.p$B[i,e])
    }
    aux <- aux + zeta.p[,,i] - zeta.p.p[,,i]
    all.prices[,,i] <- aux + zeta.p.p.p[,,i]
    all.rates[,,i] <- -1/i * log(all.prices[,,i])
  }
  
  if(indic.compute.deriv==1){
    # In this case, computation of the gradient of prices and rates w.r.t. the first column of W
    grad.prices <- array(NaN,c(n.e,n.w,max.H))
    grad.rates  <- array(NaN,c(n.e,n.w,max.H))
    aux <- 0
    g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
    g.zeta.p.p   <- matrix(NaN,n.e,n.w)
    g.zeta.p     <- matrix(NaN,n.e,n.w)
    for(i in 1:max.H){
      for(e in 1:n.e){
        g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e,1,i]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
        g.zeta.p.p[e,]   <- c(zeta.p.p[e,1,i])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
        g.zeta.p[e,]     <- c(zeta.p[e,1,i])     * (A.B.zeta.p$A[,e,i]     - xi.1)
      }
      aux <- aux + g.zeta.p - g.zeta.p.p
      grad.prices[,,i] <- aux + g.zeta.p.p.p
      grad.rates[,,i] <- -1/i * grad.prices[,,i] / all.prices[,1,i]
    }
  }else{
    grad.prices <- NaN
    grad.rates  <- NaN
  }
  return(list(prices=all.prices,rates=all.rates,grad.prices=grad.prices,grad.rates=grad.rates))
}




pricing.CDS <- function(Model,u,max.H,W,indic.compute.deriv=0,indic.delta.s=1){
  # This function computes prices and rates of bonds under the RMV assumptions
  # u has to -> - infty
  # if indic.delta.s == 0 then s_t (and not Delta s_t) is affine in w_t.
  
  n.e <- Model$n.e
  n.y <- Model$n.y
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  if(is.null(Model$a.w.s)){# No FX rate
    a.w.s <- matrix(0,n.y+n.e,n.e)
  }else{
    a.w.s <- matrix(Model$a.w.s,n.y+n.e,n.e)
  }
  if(is.null(Model$a.0.s)){# No FX rate
    a.0.s <- 0
  }else{
    a.0.s <- Model$a.0.s
  }
  
  # for zeta''':
  u.1 <- rbind(
    matrix(0,n.y,n.e),
    u * diag(n.e)
  )
  u.2 <- u.1 + a.w.s*indic.delta.s - matrix(xi.1,n.y+n.e,n.e)
  u.1 <- u.1 + a.w.s # add potential FX effect
  U <- array(0,c(n.y+n.e,n.e,2))
  U[,,1] <- u.1
  U[,,2] <- u.2
  
  A.B.zeta.p.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  # for zeta'':
  u.1 <- u.1 - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                             mu.y,mu.e,
                             alpha.y,alpha.e,
                             nu.y,
                             U,max.H)
  
  # for zeta':
  u.1 <- - a.1
  u.1 <- u.1 + a.w.s # add potential FX effect
  U[,,1] <- u.1
  
  A.B.zeta.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                           mu.y,mu.e,
                           alpha.y,alpha.e,
                           nu.y,
                           U,max.H)
  
  # for zeta tilde:
  u.1 <- 0*u.1
  u.1 <- u.1 + a.w.s # add potential FX effect
  U[,,1] <- u.1
  
  A.B.zeta.tilde <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  
  zeta.tilde <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p.p <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p   <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p     <- array(NaN,c(n.e,dim(W)[2],max.H))
  numerat <- 0
  denumerat <- 0
  for(i in 1:max.H){
    
    multipl.a.0.s <- ifelse(indic.delta.s==1,i,1)
    
    for(e in 1:n.e){
      zeta.tilde[e,,i] <- exp(          - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.tilde$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.tilde$B[i,e])
      zeta.p.p.p[e,,i] <- exp(          - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e,,i]   <- exp( - a.0[e] - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% W + A.B.zeta.p.p$B[i,e])
      zeta.p[e,,i]     <- exp( - a.0[e] - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% W + A.B.zeta.p$B[i,e])
    }
    numerat <- numerat + zeta.tilde[,,i] - zeta.p.p.p[,,i] + zeta.p.p[,,i] - zeta.p[,,i]
    denumerat <- denumerat + zeta.p.p.p[,,i]
    all.prices[,,i] <- numerat/denumerat
  }
  
  if(indic.compute.deriv==1){
    # In this case, computation of the gradient of prices and rates w.r.t. the first column of W
    grad.prices <- array(NaN,c(n.e,n.w,max.H))
    d.numer <- 0
    d.denum <- 0
    numerat.1 <- 0 # just for the first date
    denumerat.1 <- 0 # just for the first date
    g.zeta.tilde <- matrix(NaN,n.e,n.w)
    g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
    g.zeta.p.p   <- matrix(NaN,n.e,n.w)
    g.zeta.p     <- matrix(NaN,n.e,n.w)
    
    for(i in 1:max.H){
      for(e in 1:n.e){
        g.zeta.tilde[e,] <- c(zeta.tilde[e,1,i]) * (A.B.zeta.tilde$A[,e,i] - xi.1)
        g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e,1,i]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
        g.zeta.p.p[e,]   <- c(zeta.p.p[e,1,i])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
        g.zeta.p[e,]     <- c(zeta.p[e,1,i])     * (A.B.zeta.p$A[,e,i]     - xi.1)
      }
      numerat.1 <- numerat.1 + zeta.tilde[,1,i] - zeta.p.p.p[,1,i] + zeta.p.p[,1,i] - zeta.p[,1,i]
      denumerat.1 <- denumerat.1 + zeta.p.p.p[,1,i]
      d.numer <- d.numer + g.zeta.tilde - g.zeta.p.p.p + g.zeta.p.p - g.zeta.p
      d.denum <- d.denum + g.zeta.p.p.p
      grad.prices[,,i] <- (d.numer * denumerat.1 - d.denum * numerat.1) / denumerat.1^2
    }
  }else{
    grad.prices <- NaN
  }
  
  return(list(prices=all.prices,grad.prices=grad.prices,
              A.B.zeta.tilde=A.B.zeta.tilde,
              A.B.zeta.p.p.p=A.B.zeta.p.p.p,
              A.B.zeta.p.p=A.B.zeta.p.p,
              A.B.zeta.p=A.B.zeta.p,
              Model=Model,
              max.H=max.H))
}

pricing.CDS.single.date <- function(Model,u,max.H,Wt,
                                    A.B.zeta.tilde,A.B.zeta.p.p.p,A.B.zeta.p.p,A.B.zeta.p,
                                    fast.factor=1){
  # This function computes CDS prices under RFV for one date
  # Hence Wt is a column vector
  # It is used in the Extended KF
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  if(is.null(Model$a.w.s)){# No FX rate
    a.w.s <- matrix(0,n.y+n.e,n.e)
  }else{
    a.w.s <- matrix(Model$a.w.s,n.y+n.e,n.e)
  }
  if(is.null(Model$a.0.s)){# No FX rate
    a.0.s <- 0
  }else{
    a.0.s <- Model$a.0.s
  }
  
  all.prices <- matrix(NaN,n.e,max.H/fast.factor)
  
  zeta.tilde <- rep(NaN,n.e)
  zeta.p.p.p <- rep(NaN,n.e)
  zeta.p.p   <- rep(NaN,n.e)
  zeta.p     <- rep(NaN,n.e)
  
  numerat <- 0
  denumerat <- 0
  
  grad.prices <- array(NaN,c(n.e,n.w,max.H/fast.factor))
  d.numer <- 0
  d.denum <- 0
  numerat.1 <- 0 # just for the first date
  denumerat.1 <- 0 # just for the first date
  g.zeta.tilde <- matrix(NaN,n.e,n.w)
  g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
  g.zeta.p.p   <- matrix(NaN,n.e,n.w)
  g.zeta.p     <- matrix(NaN,n.e,n.w)
  
  for(i in seq(ifelse(fast.factor==1,1,fast.factor),max.H,by=fast.factor)){
    for(e in 1:n.e){
      zeta.tilde[e] <- exp(          - i*xi.0 + a.0.s + matrix(A.B.zeta.tilde$A[,e,i] - xi.1,nrow=1) %*% Wt + A.B.zeta.tilde$B[i,e])
      zeta.p.p.p[e] <- exp(          - i*xi.0 + a.0.s + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% Wt + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e]   <- exp( - a.0[e] - i*xi.0 + a.0.s + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% Wt + A.B.zeta.p.p$B[i,e])
      zeta.p[e]     <- exp( - a.0[e] - i*xi.0 + a.0.s + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% Wt + A.B.zeta.p$B[i,e])
      
      g.zeta.tilde[e,] <- c(zeta.tilde[e]) * (A.B.zeta.tilde$A[,e,i] - xi.1)
      g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
      g.zeta.p.p[e,]   <- c(zeta.p.p[e])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
      g.zeta.p[e,]     <- c(zeta.p[e])     * (A.B.zeta.p$A[,e,i]     - xi.1)
    }
    numerat <- numerat + fast.factor*(zeta.tilde - zeta.p.p.p + zeta.p.p - zeta.p)
    denumerat <- denumerat + fast.factor*(zeta.p.p.p)
    all.prices[,i/fast.factor] <- numerat/denumerat
    
    d.numer <- d.numer + fast.factor*(g.zeta.tilde - g.zeta.p.p.p + g.zeta.p.p - g.zeta.p)
    d.denum <- d.denum + fast.factor*(g.zeta.p.p.p)
    grad.prices[,,i/fast.factor] <- (d.numer * denumerat - d.denum * numerat) / denumerat^2
  }
  
  return(list(prices=all.prices,
              grad.prices=grad.prices))
}



make.Model.Q <- function(Model.P){
  
  theta.y <- Model.P$theta.priceofrisk.y
  theta.e <- Model.P$theta.priceofrisk.delta
  
  n.e <- length(Model.P$alpha.e)
  n.y <- length(Model.P$alpha.y)
  
  # The remaining part of the code comes from ZARG
  alpha.y <- Model.P$alpha.y
  alpha.e <- Model.P$alpha.e
  mu.y <- Model.P$mu.y
  mu.e <- Model.P$mu.e
  beta.yy <- Model.P$beta.yy
  beta.ey <- Model.P$beta.ey
  beta.ee <- Model.P$beta.ee
  beta.ye <- Model.P$beta.ye
  
  theta.e.tilde <- theta.e
  theta.y.tilde <- theta.y + a.function(beta.ey,mu.e,matrix(theta.e,ncol=1))
  
  # Compute parameterization for RN dynamics:
  aux <- matrix(1/(1 - theta.y.tilde * mu.y),ncol=1) %*% matrix(rep(1,n.y),nrow=1)
  beta.yy.tilde <- aux * beta.yy
  aux <- matrix(1/(1 - theta.y.tilde * mu.y),ncol=1) %*% matrix(rep(1,n.e),nrow=1)
  beta.ye.tilde <- aux * beta.ye
  aux <- matrix(1/(1 - theta.e.tilde * mu.e),ncol=1) %*% matrix(rep(1,n.y),nrow=1)
  beta.ey.tilde <- aux * beta.ey
  aux <- matrix(1/(1 - theta.e.tilde * mu.e),ncol=1) %*% matrix(rep(1,n.e),nrow=1)
  beta.ee.tilde <- aux * beta.ee
  
  mu.y.tilde <- mu.y/(1 - theta.y.tilde * mu.y)
  mu.e.tilde <- mu.e/(1 - theta.e.tilde * mu.e)
  
  alpha.y.tilde <- alpha.y/(1 - theta.y.tilde * alpha.y)
  alpha.e.tilde <- alpha.e/(1 - theta.e.tilde * alpha.e)
  
  Model.Q <- Model.P
  
  Model.Q$beta.yy <- beta.yy.tilde
  Model.Q$beta.ye <- beta.ye.tilde
  Model.Q$beta.ey <- beta.ey.tilde
  Model.Q$beta.ee <- beta.ee.tilde
  
  Model.Q$mu.y <- mu.y.tilde
  Model.Q$mu.e <- mu.e.tilde
  
  Model.Q$alpha.y <- alpha.y.tilde
  Model.Q$alpha.e <- alpha.e.tilde
  
  return(Model.Q)
}





# ===============================================
# ======= OTHER PROCEDURES ======================
# ===============================================



round.fixed.length <- function(X,n){
  # This procedure is used in the automatic creation of Latex tables
  # x is numeric. The output is a string with n numbers after ".", even if they are 0s.
  signX <- sign(X)
  if(signX>=0){
    signX <- NULL
  }else{
    signX <- "-"
  }
  x <- abs(X)
  rounded.abs.X <- round(X,n)
  string.rounded.abs.X <-toString(rounded.abs.X)
  nb.char.string.rounded.abs.X <- nchar(string.rounded.abs.X)
  
  integer.rounded.abs.X <- toString(as.integer(rounded.abs.X))
  nb.char.integer.rounded.abs.X <- nchar(integer.rounded.abs.X)
  
  string.x <- paste(signX,integer.rounded.abs.X,sep="")
  
  if(n>0){
    string.x <- paste(string.x,".",sep="")
    nb.of.available.decimal <- nb.char.string.rounded.abs.X - nb.char.integer.rounded.abs.X - 1
    for(i in 1:n){
      if(i<=nb.of.available.decimal){
        string.x <- paste(string.x,str_sub(string.rounded.abs.X,
                                           nb.char.integer.rounded.abs.X+1+i,
                                           nb.char.integer.rounded.abs.X+1+i),sep="")
      }else{
        string.x <- paste(string.x,"0",sep="")
      }
    }
    
    #string.x <- paste(signX,string.integer,".",str_replace(string.decimal,"0.",""),sep="")
  }
  return(string.x)
}



autocov <- function(X,n){
  # X is a Txk matrix
  # returns an estimator of E(X_{t-n}X_t)
  T <- dim(X)[1]
  k <- dim(X)[2]
  vec.1 <- matrix(1,1,k)
  mean.X <- apply(X,2,mean)
  X.1 <- X[1:(T-n),] - t(matrix(mean.X,k,T-n))
  X.2 <- X[(n+1):T,] - t(matrix(mean.X,k,T-n))
  return(
    matrix(1/T * apply((X.1 %x% vec.1) * (vec.1 %x% X.2),2,sum),k,k)
  )
}

NW.kappa <- function(q){
  if(q==0){
    kappa <- 0
  }else{
    kappa <- 1 - (1:q)/(q+1)
  }
  return(kappa)
}


NW.LongRunVariance <- function(X,q){
  gamma0 <- autocov(X,0)
  LRV <- gamma0
  if(q>0){
    kappa <- NW.kappa(q)
    for(i in 1:q){
      LRV <- LRV + kappa[i] * (autocov(X,i) + t(autocov(X,i)))
    }
  }
  return(LRV)
}




remove.illiquid.data <- function(X,q){
  # remove entries of X that are followed by q-1 times the same entry 
  # X has to be a vector
  T <- length(X)
  indic <- matrix(1,T-q+1,1)
  AUX <- X[1:(T-q+1)]
  for(i in 2:q){
    AUX <- cbind(AUX,X[i:(T-q+i)])
    indic <- indic * (AUX[,i]==AUX[,i-1])
  }
  indic.ext <- matrix(0,T,1)
  for(i in 2:q){
    indic.ext[i:(T-q+i)] <- indic.ext[i:(T-q+i)] + indic
  }
  X.modif <- X
  X.modif[indic.ext>0] <- NaN
  return(X.modif)
}

compute.covar.with.na <- function(X){
  T <- dim(X)[1]
  n <- dim(X)[2]
  M <- matrix(0,n,n)
  for(i in 1:n){
    for(j in 1:i){
      X.i <- X[,i]
      X.j <- X[,j]
      aux <- X.i*X.j
      indic.not.na <- which(!is.na(aux))
      X.i <- X.i[indic.not.na]
      X.j <- X.j[indic.not.na]
      M[i,j] <- cov(X.i,X.j)
    }
  }
  save.diag <- diag(M)
  M <- M + t(M)
  diag(M) <- save.diag
  return(M)
}



