####################################################
#
# Affine Modelling of Credit Risk, Pricing 
# of Credit Events and Contagion
#
# Monfort, Pegoraro, Renne & Roussellet, 2019
#
####################################################
##### Using pricing formula -- Demo codes ##########
####################################################

# The model is the one presented in Subsection 5.3 of
#  http://www.greta.it/credit/credit2017/PAPERS/Thursday/01_Monfort_Pegoraro_Renne_Roussellet.pdf




# Clear environment and console
rm(list = ls(all = TRUE)) # clear environment 

# Select working directory:
setwd("~/Dropbox/Affine_Credit/Rcode_DEMO")

# Load pricing procedures:
source("set.of.procedures.R")

par(plt=c(.1,.9,.1,.9))

# Indicate if figures have to be produced (pprint==1):
pprint <- 1



# ==========================================  
# Model specification:
# ==========================================  

# Number of entities (hence dimension of delta):
n.e <- 3
# Number of elements in y:
n.y <- n.e + 1 + n.e # in the context of this example: first n.e factors = affected by defaults;
#   next factor = common factor; others factors = entity-specific
# Dimension of w:
n.w <- n.y + n.e
# (w_t = [y_t',delta_t']')

# Specification of the short-term rate:
xi.0 <- 0
xi.1 <- rep(0,n.w)

# Specification of the recovery rate:
# ( RR_t = exp(- a.0 - a.1'w_t) )
a.1 <- matrix(rep(0,n.w),n.y+n.e,n.e) # one column = one entity
a.0 <- rep(10,n.e)

# Specifcation of y_t:
alpha.y <- c(rep(0,n.e),0,rep(.005,n.e))
beta.yy <- diag(c(rep(90,n.e),.97,rep(.9,n.e)))
beta.yy[(n.e+2):n.y,n.e+1] <- .5

beta.ye <- matrix(0,n.y,n.e) # this is beta_{e}^{(y)}, effect of delta on y
mu.y <- c(rep(.01,n.e),1,rep(1,n.e))
beta.ye[1:n.e,] <- 1/mu.y[1]
nu.y <- c(rep(0,n.e),.1,rep(0,n.e))

# Specifcation of delta_t:
alpha.e <- rep(0,n.e)
beta.ey <- cbind(.005*diag(n.e),rep(0,n.e),.00004*diag(n.e)) # this is beta_{y}^{(e)}, effect of y on delta
beta.ee <- matrix(0,n.e,n.e)
mu.e <- rep(1,n.e)


# Build a list that contains the parameterization of the physical dynamics:
Model.P <- list(
  xi.0=xi.0,xi.1=xi.1,
  a.0=a.0,a.1=a.1,
  alpha.y=alpha.y,alpha.e=alpha.e,
  mu.y=mu.y,mu.e=mu.e,
  nu.y=nu.y,
  beta.yy=beta.yy,beta.ey=beta.ey,beta.ee=beta.ee,beta.ye=beta.ye,
  n.y = dim(beta.yy)[1],
  n.e = dim(beta.ee)[1]
)

# sdf specification:
theta.y <- 0.1/mu.y
theta.e <- .7/mu.e

# Deduce Risk-neutral dynamics from physical dynamics and SDF:
Model.Q <- make.Model.Q(Model.P,theta.y,theta.e)
Model.Q.theta.y.0 <- make.Model.Q(Model.P,0*theta.y,theta.e) # alternative model (prices of risk associated to y's are 0)
Model.Q.theta.e.0 <- make.Model.Q(Model.P,  theta.y,0*theta.e) # alternative model (prices of risk associated to theta's are 0)



# ==========================================  
# Compute unconditional mean and variance of w_t:
# ==========================================  

unc.mom <- compute.moments(Model.P) # unconditional mean and variance of w_t



# ==========================================  
# Simulate the model:
# (initial state = unconditional mean)
# ==========================================  

X0 <- unc.mom$unc.mean
Y0 <- X0[1:n.y]
Y0[1:n.e] <- 0
Delta0 <- rep(0,n.e)

nb.sim <- 400

# The following line simulates a sequence for Y and Delta:
res.sim <- simul.model(alpha.y,alpha.e,
                       beta.yy,beta.ye,beta.ey,beta.ee,
                       mu.y,mu.e,
                       nu.y,
                       nb.sim,
                       Y0,Delta0)

W <- rbind(res.sim$Y,res.sim$Delta) # W is of dimension (n.y + n.e) x nb.sim

RR <- exp( -a.0[1] - a.1[,1] %*% W)




# ==========================================  
# CDS Pricing:
# ==========================================  

u <- -10000 # This u has to be large, it is the one that should go to infinity in the CDS formulas
max.H <- 20 # Maximum maturity considered for CDS

cds.pricing.P <- pricing.CDS(Model.P,u,
                             max.H,W,indic.compute.deriv = 1)
cds.pricing.Q <- pricing.CDS(Model.Q,u,
                             max.H,W,indic.compute.deriv = 1)

cds.pricing <- cds.pricing.Q

# cds.pricing is a list.
# Two main outputs:
#      => prices: array of dimension (n.e x nb.sim x max.H), where nb.sim is the number of considered periods (i.e. number of columns of W)
#      => grad.prices is computed if indic.compute.deriv == 1
#            It corresponds to the gradient of CDS w.r.t. to W_t
#            Beware: it is computed only for the first column of W, i.e. for t=1
#            The dimension of grad.prices is n.e x (n.y + n.e) x max.H

# Note:
# There is an alternative function, pricing.CDS.single.date,
#   that basically does the same. The latter function however offers the possibility 
#   to compute CDSs that do not involve payoffs at each period, but only every X periods.
#   It is valid for a single date, so W has to be of dimension (n.y + n.e) x 1
#   It has bee specifically written so as to be incorporated in the Extended Kalman Filter.

# Select maturities that will be used in plots:
matur.4.plots <- c(2,5)

if(pprint==1){
  FILE = "/figures/Figure_simulcontagion.pdf"
  pdf(file=paste(getwd(),FILE,sep=""),pointsize=10,width=8, height=8)  
}

default.date.entity.1 <- which(res.sim$Delta[1,]>0)[1]

par(mfrow=c(3,2))
par(plt=c(.12,.95,.13,.8))

plot(W[n.e+1,],type="l",lwd=2,
     main=expression(paste("(a) Common factor ",z[t],sep="")),
     xlab="",ylab="")

y2.factors <- W[1:3,]
plot(W[1,],type="l",lwd=2,
     main=expression(paste("(b) Entity-specific factors ",x[i*","*t]^{(1)},sep="")),
     xlab="",ylab="",
     ylim=c(0,max(y2.factors)),col="black")
lines(W[2,],col="dark grey",lwd=2)
lines(W[3,],col="black",lwd=2,lty=5)

legend("topright", # places a legend at the appropriate place
       c("entity 1", "entity 2", "entity 3"),
       lty=c(1,1,5), # gives the legend appropriate symbols (lines)       
       lwd=c(2,2,2), # line width
       col=c("black","dark grey","black"), # gives the legend lines the correct color and width
       seg.len = 4,bg="white"
)

y1.factors <- W[5:7,]
plot(W[5,],type="l",lwd=2,
     main=expression(paste("(c) Entity-specific factors ",x[i*","*t]^{(2)},sep="")),
     xlab="",ylab="",
     ylim=c(0,max(y1.factors)),col="black")
lines(W[6,],col="dark grey",lwd=2)
lines(W[7,],col="black",lwd=2,lty=5)

legend("topright", # places a legend at the appropriate place
       c("entity 1", "entity 2", "entity 3"),
       lty=c(1,1,5), # gives the legend appropriate symbols (lines)       
       lwd=c(2,2,2), # line width
       col=c("black","dark grey","black"), # gives the legend lines the correct color and width
       seg.len = 4,bg="white"
)



delta.factors <- W[8:10,]
plot(W[8,],type="l",lwd=2,
     main=expression(paste("(d) Credit-event variables ",delta[i*","*t],sep="")),
     xlab="",ylab="",
     ylim=c(0,max(delta.factors)),col="black")
lines(W[9,],col="dark grey",lwd=2)
lines(W[10,],col="black",lwd=2,lty=5)

legend("topright", # places a legend at the appropriate place
       c("entity 1", "entity 2", "entity 3"),
       lty=c(1,1,5), # gives the legend appropriate symbols (lines)       
       lwd=c(2,2,2), # line width
       col=c("black","dark grey","black"), # gives the legend lines the correct color and width
       seg.len = 4,bg="white"
)


cds.prem <- 10000 * cds.pricing$prices[,,matur.4.plots[1]]
plot(cds.prem[1,],type="l",lwd=2,
     main=expression(paste("(e) 2-year CDS premiums",sep="")),
     xlab="",ylab="",
     ylim=c(0,max(cds.prem,na.rm = TRUE)),col="black")
lines(cds.prem[2,],col="dark grey",lwd=2)
lines(cds.prem[3,],col="black",lwd=2,lty=5)

legend("topright", # places a legend at the appropriate place
       c("entity 1", "entity 2", "entity 3"),
       lty=c(1,1,5), # gives the legend appropriate symbols (lines)       
       lwd=c(2,2,2), # line width
       col=c("black","dark grey","black"), # gives the legend lines the correct color and width
       seg.len = 4,bg="white"
)


cds.prem <- 10000 * cds.pricing$prices[,,matur.4.plots[2]]
plot(cds.prem[1,],type="l",lwd=2,
     main=expression(paste("(f) 5-year CDS premiums",sep="")),
     xlab="",ylab="",
     ylim=c(0,max(cds.prem,na.rm = TRUE)),col="black")
lines(cds.prem[2,],col="dark grey",lwd=2)
lines(cds.prem[3,],col="black",lwd=2,lty=5)

legend("topright", # places a legend at the appropriate place
       c("entity 1", "entity 2", "entity 3"),
       lty=c(1,1,5), # gives the legend appropriate symbols (lines)       
       lwd=c(2,2,2), # line width
       col=c("black","dark grey","black"), # gives the legend lines the correct color and width
       seg.len = 4
)


if(pprint==1){
  dev.off()
}




# ==========================================  
# Compute default probabilities:
# ==========================================  

max.H <- 10

proba.def <- compute.proba.def(Model.P,max.H,W) # dimension: ne x nb.sim x max.H, where nb.sim is the number of simulation periods (number of columns of W)

matur.4.plots <- c(1,2,5)

if(pprint==1){
  FILE = "/figures/Figure_ProbaDef.pdf"
  pdf(file=paste(getwd(),FILE,sep=""),pointsize=10,width=8, height=6)  
}
par(mfrow=c(1,n.e))
for(e in 1:n.e){
  plot(proba.def[e,,matur.4.plots[1]],type="l",
       ylim=c(0,max(proba.def[e,,matur.4.plots])),
       main=paste("Proba. Def., Entity ",toString(e),sep=""))
  for(j in 2:length(matur.4.plots)){
    lines(proba.def[e,,matur.4.plots[j]],col=j)
  }
  if(e==1){
    legend("topright", # places a legend at the appropriate place
           c(
             paste("maturity = ",toString(matur.4.plots[1])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[2])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[3])," periods",sep="")),
           lty=1, # gives the legend appropriate symbols (lines)       
           lwd=1, # line width
           col=1:length(matur.4.plots), # gives the legend lines the correct color and width
           seg.len = 4,bg="white"
    )
  }
}
if(pprint==1){
  dev.off()
}



# ==========================================  
# Compute bond prices under RMV:
# ==========================================  

rmv.pricing.P <- pricing.RMV(Model.P,max.H,W)
rmv.pricing.Q <- pricing.RMV(Model.Q,max.H,W)

if(pprint==1){
  FILE = "/figures/Figure_RMV_QvsP.pdf"
  pdf(file=paste(getwd(),FILE,sep=""),pointsize=10,width=8, height=6)  
}
par(mfrow=c(1,n.e))
for(e in 1:n.e){
  plot(rmv.pricing.Q$rates[e,,1],type="l",
       ylim=c(0,max(rmv.pricing.Q$rates[e,,matur.4.plots])),
       main=paste("RMV Bond Rate, Entity ",toString(e),sep=""),lwd=2)
  lines(rmv.pricing.P$rates[e,,matur.4.plots[1]],lty=3,lwd=2)
  for(j in 2:length(matur.4.plots)){
    lines(rmv.pricing.Q$rates[e,,matur.4.plots[j]],col=j,lwd=2)
    lines(rmv.pricing.P$rates[e,,matur.4.plots[j]],col=j,lty=3,lwd=2)
  }
  if(e==1){
    legend("topright", # places a legend at the appropriate place
           c(
             paste("maturity = ",toString(matur.4.plots[1])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[2])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[3])," periods",sep="")),
           lty=1, # gives the legend appropriate symbols (lines)       
           lwd=1, # line width
           col=1:length(matur.4.plots), # gives the legend lines the correct color and width
           seg.len = 4,bg="white"
    )
  }
}
if(pprint==1){
  dev.off()
}



# ==========================================  
# Compute bond prices under RFV:
# ==========================================  


rfv.pricing.P <- pricing.RFV(Model.P,u,max.H,W,indic.compute.deriv=0)
rfv.pricing.Q <- pricing.RFV(Model.Q,u,max.H,W,indic.compute.deriv=0)
# Note: The computaion of the derivative is also available for this pricing function.

if(pprint==1){
  FILE = "/figures/Figure_RFV_QvsP.pdf"
  pdf(file=paste(getwd(),FILE,sep=""),pointsize=10,width=8, height=6)  
}
par(mfrow=c(1,n.e))
for(e in 1:n.e){
  plot(rmv.pricing.Q$rates[e,,matur.4.plots[1]],type="l",
       ylim=c(0,max(rfv.pricing.Q$rates[e,,matur.4.plots])),
       main=paste("RFV Bond Rate, Entity ",toString(e),sep=""),lwd=2)
  lines(rfv.pricing.P$rates[e,,1],lty=3,lwd=2)
  for(j in 2:length(matur.4.plots)){
    lines(rfv.pricing.Q$rates[e,,matur.4.plots[j]],col=j,lwd=2)
    lines(rfv.pricing.P$rates[e,,matur.4.plots[j]],col=j,lty=3,lwd=2)
  }
  
  if(e==1){
    legend("topright", # places a legend at the appropriate place
           c(
             paste("maturity = ",toString(matur.4.plots[1])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[2])," periods",sep=""),
             paste("maturity = ",toString(matur.4.plots[3])," periods",sep="")),
           lty=1, # gives the legend appropriate symbols (lines)       
           lwd=1, # line width
           col=1:length(matur.4.plots), # gives the legend lines the correct color and width
           seg.len = 4,bg="white"
    )
  }
  
}

if(pprint==1){
  dev.off()
}


