

simul.model <- function(alpha.y,alpha.e,
                        beta.yy,beta.ye,beta.ey,beta.ee,
                        mu.y,mu.e,
                        nu.y,
                        nb.sim,
                        Y0,Delta0,
                        use.external=FALSE,
                        external.matrix=NaN){
  
  # If use.external==TRUE then external.matrix will be used and the non-NaN elements of the latter matrix
  # will be used instead of using simulated ones.
  # external.matrix has to be of dimension (n.y+n.e)xnb.sim
  
  
  # Taking sizes
  n.y <- length(alpha.y)
  n.e <- length(alpha.e)
  
  if(use.external){
    if((dim(external.matrix)[1]!=(n.y+n.e))|(dim(external.matrix)[2]!=nb.sim)){
      print("external.matrix has not the correct dimension (should be (n.y+n.e)xnb.sim)")
      return(NaN)
    }
  }
  
  # Initialization
  Y <- matrix(0, nrow = n.y, ncol = nb.sim)
  Delta <- matrix(0, nrow = n.e, ncol = nb.sim)
  
  # Simulations of factors
  if(use.external){
    aux <- external.matrix[1:n.y,1]
    aux[is.na(aux)] <- Y0[is.na(aux)]
    Y[,1] <- aux
    
    aux <- external.matrix[(n.y+1):(n.y+n.e),1]
    aux[is.na(aux)] <- Delta0[is.na(aux)]
    Delta[,1] <- aux
  }else{
    Y[,1] <- c(Y0)
    Delta[,1] <- c(Delta0)
  }
  for (i in 2:nb.sim){
    if(use.external){
      param.pois.y <- rpois(n.y, alpha.y + beta.yy%*%Y[,(i-1)] + beta.ye%*%Delta[,(i-1)])
      aux.Y <- rgamma(n.y,shape = param.pois.y + nu.y, scale = mu.y)
      aux <- external.matrix[1:n.y,i]
      aux[is.na(aux)] <- aux.Y[is.na(aux)]
      Y[,i] <- aux
      
      param.pois.e <- rpois(n.e, alpha.e + beta.ey%*%Y[,i] + beta.ee%*%Delta[,(i-1)])
      aux.Delta <- rgamma(n.e,shape = param.pois.e, scale = mu.e)
      aux <- external.matrix[(n.y+1):(n.y+n.e),i]
      aux[is.na(aux)] <- aux.Delta[is.na(aux)]
      Delta[,i] <- aux
    }else{
      param.pois.y <- rpois(n.y, alpha.y + beta.yy%*%Y[,(i-1)] + beta.ye%*%Delta[,(i-1)])
      Y[,i] <- rgamma(n.y,shape = param.pois.y + nu.y, scale = mu.y)
      
      param.pois.e <- rpois(n.e, alpha.e + beta.ey%*%Y[,i] + beta.ee%*%Delta[,(i-1)])
      Delta[,i] <- rgamma(n.e,shape = param.pois.e, scale = mu.e)
    }
  }
  return(list(Y=Y,Delta=Delta))
}



compute.moments <- function(Model,indic.compute.uncvar=0){
  # This function computes M0 and M1 that are such that:
  #  E([y_t',delta_t']|F_{t-1}) = M0 + M1 [y_{t-1}',delta_{t-1}']
  
  n.y <- Model$n.y
  n.e <- Model$n.e
  
  M0 <- matrix(c(Model$mu.y,Model$mu.e),ncol=1) * rbind(
    matrix(Model$nu.y + Model$alpha.y,ncol=1),
    matrix(Model$alpha.e + Model$beta.ey %*% (Model$mu.y * (Model$nu.y + Model$alpha.y)))
  )
  
  M1 <- diag(c(Model$mu.y,Model$mu.e)) %*% rbind(
    cbind(Model$beta.yy,Model$beta.ye),
    cbind(
      Model$beta.ey %*% (matrix(Model$mu.y,n.y,n.y) * Model$beta.yy),
      Model$beta.ey %*% (matrix(Model$mu.y,n.y,n.e) * Model$beta.ye) + Model$beta.ee
    )
  )
  
  unc.mean <- solve(diag(n.e+n.y) - M1) %*% M0
  
  M2 <- Model$mu.y * Model$mu.y * (Model$nu.y + 2*Model$alpha.y)
  
  M3 <- 2 * diag(c(Model$mu.y^2)) %*% cbind(Model$beta.yy,Model$beta.ye)
  
  M4 <- diag(c(Model$mu.e)) %*% Model$beta.ey
  
  M5 <- 2 * Model$mu.e^2 * Model$alpha.e + 2 * diag(c(Model$mu.e^2)) %*% Model$beta.ey %*%
    (Model$mu.y * (Model$nu.y + Model$alpha.y))
  
  M6 <- 2 * diag(c(Model$mu.e^2)) %*% cbind(
    Model$beta.ey %*% diag(c(Model$mu.y)) %*% Model$beta.yy,
    Model$beta.ey %*% diag(c(Model$mu.y)) %*% Model$beta.ye + Model$beta.ee
  )
  
  Sigma.E.y.delta <- matrix(0,n.y+n.e,n.y+n.e) # This is the average conditional covariance matrix
  Sigma.E.y.delta[1:n.y,1:n.y] <- diag(c(M2 + M3 %*% unc.mean))
  Sigma.E.y.delta[(n.y+1):(n.y+n.e),(n.y+1):(n.y+n.e)] <- M4 %*% diag(c(M2 + M3 %*% unc.mean)) %*% t(M4) +
    diag(c(M5 + M6 %*% unc.mean))
  Sigma.E.y.delta[(n.y+1):(n.y+n.e),1:n.y] <- M4 %*% diag(c(M2 + M3 %*% unc.mean))
  Sigma.E.y.delta[1:n.y,(n.y+1):(n.y+n.e)] <- t(M4 %*% diag(c(M2 + M3 %*% unc.mean)))
  
  if(indic.compute.uncvar==1){
    unc.var <- solve(diag((n.y + n.e)^2) - M1 %x% M1) %*% matrix(Sigma.E.y.delta,ncol=1)
    unc.var <- matrix(unc.var,n.y + n.e,n.y + n.e)
  }else{
    unc.var <- NaN
  }
  
  return(list(
    M0 = M0,
    M1 = M1,
    M2 = M2,
    M3 = M3,
    M4 = M4,
    M5 = M5,
    M6 = M6,
    unc.mean = unc.mean,
    unc.var = unc.var
  ))
}

a.function <- function(beta,mu,u){
  # This is the function of equation 13
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # u is od dimension n.l x q
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  q <- dim(u)[2] # number of vectors u we simultaneously consider
  mu.q <- matrix(mu,ncol=1) %*% rep(1,q)
  res <- t(beta) %*% ( (u * mu.q)/(1 - u * mu.q) )
  return(res)
}

b.function <- function(beta,mu,alpha,nu,u){
  # This is the function of equation 13
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is od dimension n.l x q
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  q <- dim(u)[2] # number of vectors u we simultaneously consider
  nu.l <- dim(beta)[1]
  nu.m <- dim(beta)[2]
  mu.q <- matrix(mu,ncol=1) %*% rep(1,q)
  aux <- 1 - u * mu.q
  aux[which(nu!=0),] <- log(aux[which(nu!=0),])
  res <- matrix(alpha,nrow=1) %*% ( (u * mu.q)/(1 - u * mu.q) ) - matrix(nu,nrow=1) %*% aux
  return(res)
}

phi <- function(beta.yy,beta.ye,beta.ey,beta.ee,
                mu.y,mu.e,
                alpha.y,alpha.e,
                nu.y,
                u){
  # This is the function of equation 13
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is of dimension n.l x q x 2
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  # The first "layer" of u corresponds to u_1, the second to u_2.
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.y <- matrix(u[1:n.y,],nrow=n.y)
  u.e <- matrix(u[(n.y+1):(n.e+n.y),],nrow=n.e)
  A <- rbind(
    a.function(beta=beta.yy,mu=mu.y,u = u.y + 
                 a.function(beta=beta.ey,mu=mu.e,u = u.e)
    ),
    a.function(beta=beta.ye,mu=mu.y,u = u.y + 
                 a.function(beta=beta.ey,mu=mu.e,u = u.e)) + a.function(beta=beta.ee,mu=mu.e,u = u.e)
  )
  B <- b.function(beta=beta.ye,mu=mu.y,alpha=alpha.y,nu=nu.y,u = u.y + 
                    a.function(beta=beta.ey,mu=mu.e,u = u.e)) +
    b.function(beta=beta.ee,mu=mu.e,alpha=alpha.e,nu=0*alpha.e,u = u.e)
  return(list(A=A,B=B))
}


compute.AB <- function(beta.yy,beta.ye,beta.ey,beta.ee,
                       mu.y,mu.e,
                       alpha.y,alpha.e,
                       nu.y,
                       u,HH.aux){
  # This is the function of equation 13
  # beta is "beta_{m}^{l}", it is of dimension n.l x n.m (influence of m on l)
  # mu is of dimension n.l x 1
  # alpha is of dimension n.l x 1
  # nu is of dimension n.l x 1
  # u is of dimension n.l x q x 2
  # the function can be applied on several "u" (which correspond to the columns of matrix x)
  # The first "layer" of u corresponds to u_1, the second to u_2.
  # HH gives the considered horizons. It is either a vector or an integer that gives the maximum horizon considered
  
  if(length(HH.aux)==1){
    max.H <- HH.aux
    HH <- 1:HH.aux
  }else{
    max.H <- max(HH.aux)
    HH <- HH.aux
  }
  nb.mat <- length(HH)
  
  q <- dim(u)[2]
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  all.A <- array(NaN,c(n.y+n.e,q,nb.mat))
  all.B <- matrix(NaN,nb.mat,q)
  A <- matrix(0,n.y+n.e,q)
  B <- matrix(0,1,q)
  count <- 0
  for(i in 1:max.H){
    if(i==1){
      u.i <- u[,,1]
    }else{
      u.i <- u[,,2]
    }
    A.B <- phi(beta.yy,beta.ye,beta.ey,beta.ee,
               mu.y,mu.e,
               alpha.y,alpha.e,
               nu.y,
               u.i + A)
    A <- A.B$A
    B <- A.B$B + B
    if(sum(i==HH)==1){
      count <- count + 1
      all.A[,,count] <- A
      all.B[count,]  <- B
    }
  }
  return(list(A=all.A,B=all.B))
}


compute.proba.def <- function(Model,max.H,W){
  # This function computes the probabilities of default for the n.e entities and
  # all horizons in [1.max.H]
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- 0*Model$xi.0
  xi.1 <- 0*Model$xi.1
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.1 <- - rbind(
    matrix(0,n.y,n.e),
    100000*diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  u <- array(0,c(n.y+n.e,n.e,2))
  u[,,1] <- u.1
  u[,,2] <- u.2
  A.B <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                    mu.y,mu.e,
                    alpha.y,alpha.e,
                    nu.y,
                    u,max.H)
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  for(i in 1:max.H){
    for(e in 1:n.e){
      all.prices[e,,i] <- exp(matrix(A.B$A[,e,i] - xi.1,nrow=1) %*% W + A.B$B[i,e] - i*xi.0)
    }
  }
  
  return(1 - all.prices)
}



pricing.RMV <- function(Model,HH.aux,W){
  # This function computes prices and rates of bonds under the RMV assumptions
  
  if(length(HH.aux)==1){
    max.H <- HH.aux
    HH <- 1:HH.aux
  }else{
    max.H <- max(HH.aux)
    HH <- HH.aux
  }
  nb.mat <- length(HH)
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  u.1 <- - rbind(
    matrix(0,n.y,n.e),
    diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  u <- array(0,c(n.y+n.e,n.e,2))
  u[,,1] <- u.1
  u[,,2] <- u.2
  A.B <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                    mu.y,mu.e,
                    alpha.y,alpha.e,
                    nu.y,
                    u,HH.aux)
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],nb.mat))
  all.rates <- array(NaN,c(n.e,dim(W)[2],nb.mat))
  for(i in 1:nb.mat){
    for(e in 1:n.e){
      if(!is.null(W)){
        all.prices[e,,i] <- exp(matrix(A.B$A[,e,i] - xi.1,nrow=1) %*% W + A.B$B[i,e] - HH[i]*xi.0)
        all.rates[e,,i] <- -1/i * log(all.prices[e,,i])
      }
      A.B$A[,e,i] <- A.B$A[,e,i] - xi.1
      A.B$B[i,e]  <- A.B$B[i,e] - HH[i]*xi.0
    }
  }
  return(list(prices=all.prices,rates=all.rates,A.B=A.B))
}


pricing.RFV <- function(Model,u,max.H,W,indic.compute.deriv=0){
  # This function computes prices and rates of bonds under the RMV assumptions
  # u has to -> - infty
  # If indic.compute.deriv==1, the derivative of prices and rates will be computed
  #          w.r.t. the first column of W ONLY
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  n.w <- n.e + n.y
  
  # for zeta''':
  u.1 <- rbind(
    matrix(0,n.y,n.e),
    u * diag(n.e)
  )
  u.2 <- u.1 - matrix(xi.1,n.y+n.e,n.e)
  U <- array(0,c(n.y+n.e,n.e,2))
  U[,,1] <- u.1
  U[,,2] <- u.2
  
  A.B.zeta.p.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  # for zeta'':
  u.1 <- u.1 - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                             mu.y,mu.e,
                             alpha.y,alpha.e,
                             nu.y,
                             U,max.H)
  
  # for zeta':
  u.1 <- - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                           mu.y,mu.e,
                           alpha.y,alpha.e,
                           nu.y,
                           U,max.H)
  
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  
  zeta.p.p.p <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p   <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p     <- array(NaN,c(n.e,dim(W)[2],max.H))
  aux <- 0
  for(i in 1:max.H){
    for(e in 1:n.e){
      zeta.p.p.p[e,,i] <- exp(          - i*xi.0 + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e,,i]   <- exp( - a.0[e] - i*xi.0 + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% W + A.B.zeta.p.p$B[i,e])
      zeta.p[e,,i]     <- exp( - a.0[e] - i*xi.0 + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% W + A.B.zeta.p$B[i,e])
    }
    aux <- aux + zeta.p[,,i] - zeta.p.p[,,i]
    all.prices[,,i] <- aux + zeta.p.p.p[,,i]
    all.rates[,,i] <- -1/i * log(all.prices[,,i])
  }
  
  if(indic.compute.deriv==1){
    # In this case, computation of the gradient of prices and rates w.r.t. the first column of W
    grad.prices <- array(NaN,c(n.e,n.w,max.H))
    grad.rates  <- array(NaN,c(n.e,n.w,max.H))
    aux <- 0
    g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
    g.zeta.p.p   <- matrix(NaN,n.e,n.w)
    g.zeta.p     <- matrix(NaN,n.e,n.w)
    for(i in 1:max.H){
      for(e in 1:n.e){
        g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e,1,i]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
        g.zeta.p.p[e,]   <- c(zeta.p.p[e,1,i])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
        g.zeta.p[e,]     <- c(zeta.p[e,1,i])     * (A.B.zeta.p$A[,e,i]     - xi.1)
      }
      aux <- aux + g.zeta.p - g.zeta.p.p
      grad.prices[,,i] <- aux + g.zeta.p.p.p
      grad.rates[,,i] <- -1/i * grad.prices[,,i] / all.prices[,1,i]
    }
  }else{
    grad.prices <- NaN
    grad.rates  <- NaN
  }
  return(list(prices=all.prices,
              rates=all.rates,
              grad.prices=grad.prices,
              grad.rates=grad.rates,
              zeta.p = zeta.p,
              zeta.p.p = zeta.p.p,
              zeta.p.p.p = zeta.p.p.p))
}



pricing.CDS <- function(Model,u,max.H,W,indic.compute.deriv=0,indic.delta.s=1){
  # This function computes prices and rates of bonds under the RMV assumptions
  # u has to -> - infty
  # if indic.delta.s == 0 then s_t (and not Delta s_t) is affine in w_t.
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  if(is.null(Model$a.w.s)){# No FX rate
    a.w.s <- matrix(0,n.y+n.e,n.e)
  }else{
    a.w.s <- matrix(Model$a.w.s,n.y+n.e,n.e)
  }
  if(is.null(Model$a.0.s)){# No FX rate
    a.0.s <- 0
  }else{
    a.0.s <- Model$a.0.s
  }
  
  # for zeta''':
  u.1 <- rbind(
    matrix(0,n.y,n.e),
    u * diag(n.e)
  )
  u.2 <- u.1 + a.w.s*indic.delta.s - matrix(xi.1,n.y+n.e,n.e)
  u.1 <- u.1 + a.w.s # add potential FX effect
  U <- array(0,c(n.y+n.e,n.e,2))
  U[,,1] <- u.1
  U[,,2] <- u.2
  
  A.B.zeta.p.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  # for zeta'':
  u.1 <- u.1 - a.1
  U[,,1] <- u.1
  
  A.B.zeta.p.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                             mu.y,mu.e,
                             alpha.y,alpha.e,
                             nu.y,
                             U,max.H)
  
  # for zeta':
  u.1 <- - a.1
  u.1 <- u.1 + a.w.s # add potential FX effect
  U[,,1] <- u.1
  
  A.B.zeta.p <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                           mu.y,mu.e,
                           alpha.y,alpha.e,
                           nu.y,
                           U,max.H)
  
  # for zeta tilde:
  u.1 <- 0*u.1
  u.1 <- u.1 + a.w.s # add potential FX effect
  U[,,1] <- u.1
  
  A.B.zeta.tilde <- compute.AB(beta.yy,beta.ye,beta.ey,beta.ee,
                               mu.y,mu.e,
                               alpha.y,alpha.e,
                               nu.y,
                               U,max.H)
  
  
  all.prices <- array(NaN,c(n.e,dim(W)[2],max.H))
  all.rates <- array(NaN,c(n.e,dim(W)[2],max.H))
  
  zeta.tilde <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p.p <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p.p   <- array(NaN,c(n.e,dim(W)[2],max.H))
  zeta.p     <- array(NaN,c(n.e,dim(W)[2],max.H))
  numerat <- 0
  denumerat <- 0
  for(i in 1:max.H){
    
    multipl.a.0.s <- ifelse(indic.delta.s==1,i,1)
    
    for(e in 1:n.e){
      zeta.tilde[e,,i] <- exp(          - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.tilde$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.tilde$B[i,e])
      zeta.p.p.p[e,,i] <- exp(          - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% W + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e,,i]   <- exp( - a.0[e] - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% W + A.B.zeta.p.p$B[i,e])
      zeta.p[e,,i]     <- exp( - a.0[e] - i*xi.0 + multipl.a.0.s*a.0.s + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% W + A.B.zeta.p$B[i,e])
    }
    numerat <- numerat + zeta.tilde[,,i] - zeta.p.p.p[,,i] + zeta.p.p[,,i] - zeta.p[,,i]
    denumerat <- denumerat + zeta.p.p.p[,,i]
    all.prices[,,i] <- numerat/denumerat
  }
  
  if(indic.compute.deriv==1){
    # In this case, computation of the gradient of prices and rates w.r.t. the first column of W
    grad.prices <- array(NaN,c(n.e,n.w,max.H))
    d.numer <- 0
    d.denum <- 0
    numerat.1 <- 0 # just for the first date
    denumerat.1 <- 0 # just for the first date
    g.zeta.tilde <- matrix(NaN,n.e,n.w)
    g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
    g.zeta.p.p   <- matrix(NaN,n.e,n.w)
    g.zeta.p     <- matrix(NaN,n.e,n.w)
    
    for(i in 1:max.H){
      for(e in 1:n.e){
        g.zeta.tilde[e,] <- c(zeta.tilde[e,1,i]) * (A.B.zeta.tilde$A[,e,i] - xi.1)
        g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e,1,i]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
        g.zeta.p.p[e,]   <- c(zeta.p.p[e,1,i])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
        g.zeta.p[e,]     <- c(zeta.p[e,1,i])     * (A.B.zeta.p$A[,e,i]     - xi.1)
      }
      numerat.1 <- numerat.1 + zeta.tilde[,1,i] - zeta.p.p.p[,1,i] + zeta.p.p[,1,i] - zeta.p[,1,i]
      denumerat.1 <- denumerat.1 + zeta.p.p.p[,1,i]
      d.numer <- d.numer + g.zeta.tilde - g.zeta.p.p.p + g.zeta.p.p - g.zeta.p
      d.denum <- d.denum + g.zeta.p.p.p
      grad.prices[,,i] <- (d.numer * denumerat.1 - d.denum * numerat.1) / denumerat.1^2
    }
  }else{
    grad.prices <- NaN
  }
  
  return(list(prices=all.prices,grad.prices=grad.prices,
              A.B.zeta.tilde=A.B.zeta.tilde,
              A.B.zeta.p.p.p=A.B.zeta.p.p.p,
              A.B.zeta.p.p=A.B.zeta.p.p,
              A.B.zeta.p=A.B.zeta.p,
              Model=Model,
              max.H=max.H))
}

pricing.CDS.single.date <- function(Model,u,max.H,Wt,
                                    A.B.zeta.tilde,A.B.zeta.p.p.p,A.B.zeta.p.p,A.B.zeta.p,
                                    fast.factor=1){
  # This function computes CDS prices under RFV for one date
  # Hence Wt is a column vector
  # It is aimed at feeding the Extended KF
  
  beta.yy <- Model$beta.yy
  beta.ye <- Model$beta.ye
  beta.ey <- Model$beta.ey
  beta.ee <- Model$beta.ee
  
  mu.y <- Model$mu.y
  mu.e <- Model$mu.e
  
  alpha.y <- Model$alpha.y
  alpha.e <- Model$alpha.e
  
  nu.y <- Model$nu.y
  
  xi.0 <- Model$xi.0
  xi.1 <- Model$xi.1
  
  a.0 <- Model$a.0
  a.1 <- Model$a.1
  
  q <- dim(u)[2]
  
  if(class(beta.yy)=="numeric"){
    n.y <- 1
  }else{
    n.y <- dim(beta.yy)[1]
  }
  if(class(beta.ee)=="numeric"){
    n.e <- 1
  }else{
    n.e <- dim(beta.ee)[1]
  }
  
  if(is.null(Model$a.w.s)){# No FX rate
    a.w.s <- matrix(0,n.y+n.e,n.e)
  }else{
    a.w.s <- matrix(Model$a.w.s,n.y+n.e,n.e)
  }
  if(is.null(Model$a.0.s)){# No FX rate
    a.0.s <- 0
  }else{
    a.0.s <- Model$a.0.s
  }
  
  all.prices <- matrix(NaN,n.e,max.H/fast.factor)
  
  zeta.tilde <- rep(NaN,n.e)
  zeta.p.p.p <- rep(NaN,n.e)
  zeta.p.p   <- rep(NaN,n.e)
  zeta.p     <- rep(NaN,n.e)
  
  numerat <- 0
  denumerat <- 0
  
  grad.prices <- array(NaN,c(n.e,n.w,max.H/fast.factor))
  d.numer <- 0
  d.denum <- 0
  numerat.1 <- 0 # just for the first date
  denumerat.1 <- 0 # just for the first date
  g.zeta.tilde <- matrix(NaN,n.e,n.w)
  g.zeta.p.p.p <- matrix(NaN,n.e,n.w)
  g.zeta.p.p   <- matrix(NaN,n.e,n.w)
  g.zeta.p     <- matrix(NaN,n.e,n.w)
  
  for(i in seq(ifelse(fast.factor==1,1,fast.factor),max.H,by=fast.factor)){
    for(e in 1:n.e){
      zeta.tilde[e] <- exp(          - i*xi.0 + a.0.s + matrix(A.B.zeta.tilde$A[,e,i] - xi.1,nrow=1) %*% Wt + A.B.zeta.tilde$B[i,e])
      zeta.p.p.p[e] <- exp(          - i*xi.0 + a.0.s + matrix(A.B.zeta.p.p.p$A[,e,i] - xi.1,nrow=1) %*% Wt + A.B.zeta.p.p.p$B[i,e])
      zeta.p.p[e]   <- exp( - a.0[e] - i*xi.0 + a.0.s + matrix(A.B.zeta.p.p$A[,e,i]   - xi.1,nrow=1) %*% Wt + A.B.zeta.p.p$B[i,e])
      zeta.p[e]     <- exp( - a.0[e] - i*xi.0 + a.0.s + matrix(A.B.zeta.p$A[,e,i]     - xi.1,nrow=1) %*% Wt + A.B.zeta.p$B[i,e])
      
      g.zeta.tilde[e,] <- c(zeta.tilde[e]) * (A.B.zeta.tilde$A[,e,i] - xi.1)
      g.zeta.p.p.p[e,] <- c(zeta.p.p.p[e]) * (A.B.zeta.p.p.p$A[,e,i] - xi.1)
      g.zeta.p.p[e,]   <- c(zeta.p.p[e])   * (A.B.zeta.p.p$A[,e,i]   - xi.1)
      g.zeta.p[e,]     <- c(zeta.p[e])     * (A.B.zeta.p$A[,e,i]     - xi.1)
    }
    numerat <- numerat + fast.factor*(zeta.tilde - zeta.p.p.p + zeta.p.p - zeta.p)
    denumerat <- denumerat + fast.factor*(zeta.p.p.p)
    all.prices[,i/fast.factor] <- numerat/denumerat
    
    d.numer <- d.numer + fast.factor*(g.zeta.tilde - g.zeta.p.p.p + g.zeta.p.p - g.zeta.p)
    d.denum <- d.denum + fast.factor*(g.zeta.p.p.p)
    grad.prices[,,i/fast.factor] <- (d.numer * denumerat - d.denum * numerat) / denumerat^2
  }
  
  return(list(prices=all.prices,
              grad.prices=grad.prices))
}




make.Model.Q <- function(Model.P,theta.y,theta.e){
  
  n.e <- length(Model.P$alpha.e)
  n.y <- length(Model.P$alpha.y)
  
  # The remaining part of the code comes from ZARG
  alpha.y <- Model.P$alpha.y
  alpha.e <- Model.P$alpha.e
  mu.y <- Model.P$mu.y
  mu.e <- Model.P$mu.e
  beta.yy <- Model.P$beta.yy
  beta.ey <- Model.P$beta.ey
  beta.ee <- Model.P$beta.ee
  beta.ye <- Model.P$beta.ye
  
  theta.e.tilde <- theta.e
  theta.y.tilde <- theta.y + a.function(beta.ey,mu.e,matrix(theta.e,ncol=1))
  
  # Compute parameterization if RN dynamics:
  aux <- matrix(1/(1 - theta.y.tilde * mu.y),ncol=1) %*% matrix(rep(1,n.y),nrow=1)
  beta.yy.tilde <- aux * beta.yy
  aux <- matrix(1/(1 - theta.y.tilde * mu.y),ncol=1) %*% matrix(rep(1,n.e),nrow=1)
  beta.ye.tilde <- aux * beta.ye
  aux <- matrix(1/(1 - theta.e.tilde * mu.e),ncol=1) %*% matrix(rep(1,n.y),nrow=1)
  beta.ey.tilde <- aux * beta.ey
  aux <- matrix(1/(1 - theta.e.tilde * mu.e),ncol=1) %*% matrix(rep(1,n.e),nrow=1)
  beta.ee.tilde <- aux * beta.ee
  
  mu.y.tilde <- mu.y/(1 - theta.y.tilde * mu.y)
  mu.e.tilde <- mu.e/(1 - theta.e.tilde * mu.e)
  
  alpha.y.tilde <- alpha.y/(1 - theta.y.tilde * alpha.y)
  alpha.e.tilde <- alpha.e/(1 - theta.e.tilde * alpha.e)
  
  Model.Q <- Model.P
  
  Model.Q$xi.0 <- xi.0
  Model.Q$xi.1 <- xi.1
  
  Model.Q$beta.yy <- beta.yy.tilde
  Model.Q$beta.ye <- beta.ye.tilde
  Model.Q$beta.ey <- beta.ey.tilde
  Model.Q$beta.ee <- beta.ee.tilde
  
  Model.Q$mu.y <- mu.y.tilde
  Model.Q$mu.e <- mu.e.tilde
  
  Model.Q$alpha.y <- alpha.y.tilde
  Model.Q$alpha.e <- alpha.e.tilde
  
  return(Model.Q)
}


